# -*- coding: utf-8 -*-
#############################################################################
# SRW Samples library
# Authors/Contributors: Maksim Rakitin, Rebecca Ann Coles
# October 26, 2016
# April 11, 2020
#############################################################################

import math
import os
import sys
import time
from array import array

import srwlib
import uti_io


# ********************** The class for Samples:
class SRWLUtiSmp:
    """The class for Samples from image file (e.g., .tif), NumPy array (.npy), etc.

    :param file_path: full path to the image or the saved NumPy array.
    :param area: the coordinates of the rectangle area listed in the following order: x_start, x_end, y_start, y_end.
    :param rotate_angle: the angle [deg] to rotate the read image counterclockwise. See scipy.ndimage.interpolation.rotate() for details.
    :param rotate_reshape: if reshape is true, the output shape is adapted so that the input array is contained completely in the output.
    :param cutoff_background_noise: the ratio for cutoff the background noise (between 0 and 1).
    :param background_color: the background color code to use instead of the background noise (0=black, 255=white).
    :param tile: the list/tuple (rows, columns) to tile the cut area of the image. See numpy.tile() for details.
    :param shift_x: shift the whole image horizontally. Positive value shifts the image to the right, negative - to the left. See numpy.pad() for details.
    :param shift_y: shift the whole image vertically. Positive value shifts the image to the top, negative - to the bottom. See numpy.pad() for details.
    :param invert: invert the image. See numpy.invert() for details.
    :param is_show_images: a flag to show the initial and processed images.
    :param is_save_images: a flag to save the initial and processed images.
    :param raw_image_name: the name of the raw file in case if it's saved.
    :param processed_image_name: the name of the processed file in case if it's saved.
    :param prefix: the prefix to add to the names of the saved image files.
    :param output_image_format: the format of the output file. If not specified, the input format is used.
    """

    def __init__(self, file_path,
                 area=None, rotate_angle=0, rotate_reshape=True, cutoff_background_noise=0.25, background_color=0,
                 tile=None, shift_x=None, shift_y=None, invert=None,
                 is_show_images=False, is_save_images=False,
                 raw_image_name='raw', processed_image_name='processed', prefix='', output_image_format=None, #):
                 cutoff_max_fact=None, max_color=255): #OC29052020
        # Input parameters:
        self.file_path = file_path
        self.area = area
        self.rotate_angle = rotate_angle if rotate_angle is not None else 0
        self.rotate_reshape = True if rotate_reshape else False
        self.cutoff_background_noise = cutoff_background_noise if cutoff_background_noise is not None else 0
        self.background_color = background_color if background_color is not None else 0
        self.invert = invert
        self.tile = tile
        self.shift_x = shift_x
        self.shift_y = shift_y
        self.is_show_images = is_show_images
        self.is_save_images = is_save_images
        output_image_format = os.path.splitext(file_path)[1].replace('.', '') if not output_image_format else \
            output_image_format
        self.raw_image_name = self._add_prefix(prefix, raw_image_name, output_image_format)
        self.processed_image_name = self._add_prefix(prefix, processed_image_name, output_image_format)

        self.cutoff_max_fact = cutoff_max_fact #OC29052020
        self.max_color = max_color if max_color is not None else 255 #OC29052020

        # Output parameters:
        self.data = None
        self.raw_image = None
        self.processed_image = None
        self.nx = None
        self.ny = None
        self.limit_value = None

        # Check input type automatically:
        self.input_type = self._check_input_type(file_path)

        # Set the dir where to save the files:
        self.save_dir = os.path.abspath(os.path.dirname(file_path))

        # Check the input file(s):
        self._check_files()

        # Process input:
        self.read_sample()

        # Show the resulted images:
        if self.is_show_images:
            self.show_images()

        # Save the resulted images:
        if self.is_save_images:
            self.save_images()

    def get_data_from_image(self):
        import_err_msg = '"{}" library cannot be imported. Please install it first with "pip install {}".'
        try:
            import numpy as np
        except ImportError:
            raise ImportError(import_err_msg.format('NumPy', 'numpy'))
        try:
            from PIL import Image
        except ImportError:
            raise ImportError(import_err_msg.format('PIL', 'pillow'))
        try:
            from scipy.ndimage.interpolation import rotate
        except ImportError:
            raise ImportError(import_err_msg.format('SciPy', 'scipy'))

        d = uti_io.read_image(image_path=self.file_path)
        self.data = d['data']
        self.raw_image = d['raw_image']
        self.limit_value = d['limit_value']

        # Remove background noise:
        #OC29052020 (moved down)
        #assert 0 <= self.background_color <= 255, 'Background color ({}) should be between 0 and 255'.format(
        #    self.background_color)
        #assert 0 <= self.cutoff_background_noise <= 1, 'Cutoff background noise ({}) should be between 0 and 1'.format(
        #    self.cutoff_background_noise)
        #self.data[np.where(self.data < self.limit_value * self.cutoff_background_noise)] = np.uint16(
        #    self.background_color)

        if self.area:
            assert type(self.area) in [tuple, list] and len(self.area) == 4, \
                'The area should be a list/tuple and contain 4 elements (x_start, x_end, y_start, y_end)'
            assert self.area[0] <= self.data.shape[1] and self.area[1] <= self.data.shape[1] and \
                   self.area[2] <= self.data.shape[0] and self.area[3] <= self.data.shape[0], \
                '\n  x_start ({}) and x_end ({}) should be less than {}\n  y_start ({}) and y_end ({}) should be less than {}'.format(
                    self.area[0], self.area[1], self.data.shape[1],
                    self.area[2], self.area[3], self.data.shape[0],
                )
            assert self.area[0] < self.area[1] and \
                   self.area[2] < self.area[3], \
                '\n  x_start ({}) should be less than x_end ({})\n  y_start ({}) should be less than y_end ({})'.format(
                    self.area[0], self.area[1],
                    self.area[2], self.area[3],
                )
            self.data = self.data[self.area[2]:self.area[3], self.area[0]:self.area[1]]

        # Remove background noise:
        #OC29052020 (moved here to make operations on reduced amount of data)
        assert 0 <= self.background_color <= 255, 'Background color ({}) should be between 0 and 255'.format(
            self.background_color)
        assert 0 <= self.cutoff_background_noise <= 1, 'Cutoff background noise ({}) should be between 0 and 1'.format(
            self.cutoff_background_noise)
        self.data[np.where(self.data < self.limit_value * self.cutoff_background_noise)] = np.uint16(
            self.background_color)

        if(hasattr(self, 'cutoff_max_fact')): #OC29052020
           if(self.cutoff_max_fact is not None):
                assert 0 <= self.cutoff_max_fact <= 1, 'Maximal cutoff signal value ({}) should be between 0 and 1'.format(self.cutoff_max_fact)
                assert 0 <= self.max_color <= 255, 'Maximum color value ({}) should be between 0 and 255'.format(self.max_color)
                self.data[np.where(self.data >= self.limit_value * self.cutoff_max_fact)] = np.uint16(self.max_color)

        if self.tile:
            assert type(self.tile) in [list, tuple], 'The type of tile ({}) should be list/tuple'.format(
                type(self.tile).__name__)
            assert len(self.tile) == 2, 'The size ({}) of the list/tuple should be 2'.format(len(self.tile))
            self.data = np.tile(self.data, self.tile)

        if self.rotate_angle:
            assert -360 < self.rotate_angle < 360, 'The angle should be from -360 to 360 degrees.'
            self.data = rotate(self.data, self.rotate_angle, axes=(1, 0), reshape=self.rotate_reshape, output=None,
                               order=0, mode='constant', cval=self.background_color, prefilter=False)

        if self.shift_x:
            assert type(self.shift_x) is int, 'Type of shift_x ({}) should be int'.format(type(self.shift_x).__name__)
            if self.shift_x > 0:
                self.data = np.pad(self.data, ((0, 0), (self.shift_x, 0)), mode='constant',
                                   constant_values=(self.background_color))[:, :-self.shift_x]
            else:
                self.data = np.pad(self.data, ((0, 0), (0, -self.shift_x)), mode='constant',
                                   constant_values=(self.background_color))[:, -self.shift_x:]

        if self.shift_y:
            assert type(self.shift_y) is int, 'Type of shift_y ({}) should be int'.format(type(self.shift_y).__name__)
            if self.shift_y < 0:
                self.data = np.pad(self.data, ((-self.shift_y, 0), (0, 0)), mode='constant',
                                   constant_values=(self.background_color))[:self.shift_y, :]
            else:
                self.data = np.pad(self.data, ((0, self.shift_y), (0, 0)), mode='constant',
                                   constant_values=(self.background_color))[self.shift_y:, :]

        if self.invert:
            self.data = np.invert(self.data)

        self.processed_image = Image.fromarray(self.data)

    def get_image_from_data(self):
        data = np.load(self.file_path)
        self.limit_value = 255
        #Suggested by RAC(?):
        self.data = (np.subtract(data, data.min(), dtype=np.float32)) / (np.subtract(data.max(), data.min(), dtype=np.float32)) * self.limit_value
        #self.data = (data - data.min()) / (data.max() - data.min()) * self.limit_value
        self.raw_image = Image.fromarray(self.data.astype(np.uint8))
        self.processed_image = self.raw_image

    def read_sample(self):
        if self.input_type == 'image':
            self.get_data_from_image()
        elif self.input_type == 'npy':
            self.get_image_from_data()
        else:
            raise NotImplementedError(
                'Processing of the "{}" input type is not implemented yet.'.format(self.input_type))
        self.nx = self.data.shape[1]
        self.ny = self.data.shape[0]

    def save_images(self):
        # self.raw_image.save(os.path.join(self.save_dir, self.raw_image_name))
        self.processed_image.save(os.path.join(self.save_dir, self.processed_image_name))

    def show_images(self):
        self.raw_image.show()
        self.processed_image.show()

    def _add_prefix(self, prefix, name, image_format):
        output_name = '{}_{}'.format(prefix, name) if prefix else name
        return '{}.{}'.format(output_name, image_format)

    def _check_files(self):
        if not os.path.isfile(self.file_path):
            raise ValueError('Provided file "{}" does not exist.'.format(self.file_path))

    def _check_input_type(self, file_path):
        self.possible_extensions = {
            'image': ['tif', 'tiff', 'png', 'bmp', 'gif', 'jpg', 'jpeg'],
            'npy': ['npy'],
        }
        extension = os.path.splitext(file_path)[1][1:].lower()
        for k in self.possible_extensions.keys():
            for e in self.possible_extensions[k]:
                if extension == e:
                    return k
        all_extensions = [x for x_list in self.possible_extensions.values() for x in x_list]
        all_extensions += [x.upper() for x in all_extensions]
        raise ValueError('Incorrect extension: {}. Possible values: {}.'.format(extension, ', '.join(all_extensions)))


# ********************** Create transmission element from the data from an image file:
def srwl_opt_setup_transm_from_file(
        file_path, resolution, thickness, delta, atten_len,
        arTr=None, extTr=0, fx=1e+23, fy=1e+23,
        xc=0, yc=0, ne=1, e_start=0, e_fin=0,
        area=None, rotate_angle=None, rotate_reshape=None,
        cutoff_background_noise=None, background_color=None,
        tile=None, shift_x=None, shift_y=None, invert=None,
        is_save_images=True, prefix='', output_image_format=None,
        _cutoff_max_fact=None, _max_color=255, _ret='srw' #OC28052020
):
    """Setup Sample element.

    :param file_path: path to the input file (image or .npy).
    :param resolution: resolution of the image [m/pixel].
    :param thickness: thickness of the sample [m].
    :param delta: refractive index decrement.
    :param atten_len: attenuation length [m].
    :param arTr: complex C-aligned data array (of 2*ne*nx*ny length) storing amplitude transmission and optical path difference as function of transverse coordinates.
    :param extTr: transmission outside the grid/mesh is zero (0), or it is same as on boundary (1).
    :param fx: estimated focal length in the horizontal plane [m].
    :param fy: estimated focal length in the vertical plane [m].
    :param xc: horizontal coordinate of center [m].
    :param yc: vertical coordinate of center [m].
    :param ne: number of transmission data points vs photon energy.
    :param e_start: initial photon energy [eV].
    :param e_fin: final photon energy [eV].
    :param area: the coordinates of the rectangle area listed in the following order: x_start, x_end, y_start, y_end.
    :param rotate_angle: the angle [deg] to rotate the read image counterclockwise. See scipy.ndimage.interpolation.rotate() for details.
    :param rotate_reshape: if reshape is true, the output shape is adapted so that the input array is contained completely in the output.
    :param cutoff_background_noise: the ratio for cutoff the background noise (between 0 and 1).
    :param background_color: the background color code to use instead of the background noise (0=black, 255=white).
    :param tile: the list/tuple (rows, columns) to tile the cut area of the image. See numpy.tile() for details.
    :param shift_x: shift the whole image horizontally. Positive value shifts the image to the right, negative - to the left. See numpy.pad() for details.
    :param shift_y: shift the whole image vertically. Positive value shifts the image to the top, negative - to the bottom. See numpy.pad() for details.
    :param invert: invert the image. See numpy.invert() for details.
    :param is_save_images: a flag to save the initial and processed images.
    :param prefix: the prefix to add to the names of the saved image files.
    :param output_image_format: the format of the output file. If not specified, the input format is used.
    :param _cutoff_max_fact: the ratio/factor for maximum signal cutoff (between 0 and 1).
    :param _max_color: the maximum color code to use for signal level above the cutoff_max_fact (0=black, 255=white).
    :param _ret: type of return: 'srw' returns SRWLOptT type transmission object which simulates the Sample (default), 'img' returns processed PIL image object, 'all' returns both SRWLOptT object and PIL image.
    :return: by default, transmission (SRWLOptT) type optical element which simulates the Sample; other options are defined by _ret variable.
    """

    input_parms = {
        "type": "sample",
        "resolution": resolution,
        "thickness": thickness,
        "refractiveIndex": delta,
        "attenuationLength": atten_len,
        "horizontalCenterCoordinate": xc,
        "verticalCenterCoordinate": yc,
        "initialPhotonEnergy": e_start,
        "finalPhotonPnergy": e_fin,
        'area': area,
        'rotateAngle': rotate_angle,
        'rotateReshape': rotate_reshape,
        'cutoffBackgroundNoise': cutoff_background_noise,
        'backgroundColor': background_color,
        'tile': tile,
        'shiftX': shift_x,
        'shiftY': shift_y,
        'invert': invert,
        'outputImageFormat': output_image_format,
        'cutoffMaxFact': _cutoff_max_fact,
        'maxColor': _max_color,
    }

    s = SRWLUtiSmp(
        file_path=file_path,
        area=area,
        rotate_angle=rotate_angle,
        rotate_reshape=rotate_reshape,
        cutoff_background_noise=cutoff_background_noise,
        background_color=background_color,
        tile=tile,
        shift_x=shift_x,
        shift_y=shift_y,
        invert=invert,
        is_show_images=False,
        is_save_images=is_save_images,
        prefix=prefix,
        output_image_format=output_image_format,
        cutoff_max_fact=_cutoff_max_fact, #OC29052020
        max_color=_max_color, #OC29052020
    )

    # Input parameters to SRWLOptT:
    nx = s.nx 
    ny = s.ny
    #print(nx, ny) #DEBUG
    
    rx = nx * resolution
    ry = ny * resolution

    #opT = srwlib.SRWLOptT(_nx=nx, _ny=ny, _rx=rx, _ry=ry,
    #                      _arTr=arTr, _extTr=extTr, _Fx=fx, _Fy=fy,
    #                      _x=xc, _y=yc, _ne=ne, _eStart=e_start, _eFin=e_fin)

    data = s.data

    if(_ret == 'img'): return s.processed_image #OC28052020

    #OC10112018
    specPropAreDef = False
    if(ne > 1):
        if((isinstance(delta, list) or isinstance(delta, array)) and (isinstance(atten_len, list) or isinstance(atten_len, array))):
            lenDelta = len(delta)
            if((lenDelta == len(atten_len)) and (lenDelta == ne)): specPropAreDef = True
            else: raise Exception("Inconsistent spectral refractive index decrement and/or attenuation length data")

    #OC10112018
    useNumPy = False
    try:
        import numpy as np
        useNumPy = True
    except:
        print('NumPy can not be loaded, native Python arrays / lists will be used instead, impacting performance')

    if(useNumPy): #RC161018
        
        thickByLim = thickness/s.limit_value
        nxny = nx*ny
        miHalfThickByLimByAttenLen = None
        miThickDeltaByLim = None

        if(ne <= 1):
            miHalfThickByLimByAttenLen = -0.5*thickByLim/atten_len
            miThickDeltaByLim = -thickByLim*delta

            #amplTransm = np.exp(-0.5 * data * thickness / (s.limit_value * atten_len))
            amplTransm = np.exp(miHalfThickByLimByAttenLen*data)

            #optPathDiff =  -1 * data * thickness * delta / s.limit_value
            optPathDiff =  miThickDeltaByLim*data

            arTr = np.empty((2*nxny), dtype=float)

            #print(len(amplTransm), len(amplTransm[0])) #DEBUG
            
            arTr[0::2] = np.reshape(amplTransm, nxny)
            arTr[1::2] = np.reshape(optPathDiff, nxny)
            #opT.arTr = arTr
            
        else:
            two_ne = 2*ne
            arTr = np.empty((nxny*two_ne), dtype=float)

            if(specPropAreDef == False):
                miHalfThickByLimByAttenLen = -0.5*thickByLim/atten_len
                miThickDeltaByLim = -thickByLim*delta
            
            for ie in range(ne):
                if(specPropAreDef):
                    miHalfThickByLimByAttenLen = -0.5*thickByLim/atten_len[ie]
                    miThickDeltaByLim = -thickByLim*delta[ie]

                amplTransm = np.exp(miHalfThickByLimByAttenLen*data)
                optPathDiff = miThickDeltaByLim*data

                two_ie = 2*ie
                arTr[two_ie::two_ne] = np.reshape(amplTransm, nxny) #to check!
                arTr[(two_ie + 1)::two_ne] = np.reshape(optPathDiff, nxny)
    else:
        #Same data alignment as for wavefront: outmost loop vs y, inmost loop vs e
        nTot = 2*ne*nx*ny
        arTr = array('d', [0]*nTot)
    
        offset = 0
        for iy in range(ny):
            for ix in range(nx):
                #In images Y=0 corresponds from upper-left corner, in SRW it's lower-left corner:
                pathInBody = thickness * data[ny - iy - 1, ix] / s.limit_value
            
                #OC10112018
                miHalfPathInBody = -0.5*pathInBody
            
                if(specPropAreDef):
                    for ie in range(ne):
                        #opT.arTr[offset] = math.exp(-0.5 * pathInBody / atten_len)  # amplitude transmission
                        #opT.arTr[offset + 1] = -delta * pathInBody  # optical path difference
                        arTr[offset] = math.exp(miHalfPathInBody / atten_len[ie]) #amplitude transmission
                        arTr[offset + 1] = -delta[ie] * pathInBody #optical path difference
                        offset += 2
                else:
                    for ie in range(ne):
                        arTr[offset] = math.exp(miHalfPathInBody / atten_len) #amplitude transmission
                        arTr[offset + 1] = -delta * pathInBody  #optical path difference
                        offset += 2

    opT = srwlib.SRWLOptT(_nx=nx, _ny=ny, _rx=rx, _ry=ry,
                          _arTr=arTr, _extTr=extTr, _Fx=fx, _Fy=fy,
                          _x=xc, _y=yc, _ne=ne, _eStart=e_start, _eFin=e_fin)

    opT.input_parms = input_parms

    #OC28052020
    if(_ret == 'srw'): return opT 
    elif(_ret == 'all'): return opT, s.processed_image
    else: return None
    #return opT

# ********************** Create transmission element from the data from a generated random 2D disk
def srwl_opt_setup_smp_rnd_obj2d( #RAC06052020
        _thickness, _delta, _atten_len, _rx = 10.e-06, _ry = 10.e-06, _xc = 0, _yc = 0, _nx = 1001, _ny = 1001,
        _dens = 1, _edge_frac = 0.02, _sim_step_size = 1.e-06,
        _obj_type = 1, _r_min_bw_obj = 0.1e-06, 
        _obj_size_min = 0.1e-06, _obj_size_max = 0.1e-06, _size_dist = 1, 
        _ang_min = 0, _ang_max = 0, _ang_dist = 1, _rand_alg = 1,
        _obj_par1 = None, _obj_par2 = None,
        _ext_tr = 1, _e_start = 0, _e_fin = 0, _ret = 'srw'): #OC01062020
        #_ext_tr = 1, _ne = 1, _e_start = 0, _e_fin = 0, _ret = 'srw'): #OC24052020
        #_extTr=1, _ne=1, _e_start=0, _e_fin=0, _ret='srw', _file_path='data_example_17', _file_name='ex17_res_opt_path_dif-ASCII'):
    """Setup Transmission type Optical Element which simulates Random 2D Disk using "srwl_uti_smp_rand_obj2d.py" 

    :param _thickness: thickness of the sample [m]
    :param _delta: refractive index decrement of sample material
    :param _atten_len: attenuation lengthof sample material [m]
    :param _rx: range of the horizontal coordinate [m] for which the transmission is defined
    :param _ry: range of the vertical coordinate [m] for which the transmission is defined
    :param _xc: horizontal coordinate of center [m]
    :param _yc: vertical coordinate of center [m]
    :param _nx: number of transmission data points in the horizontal direction
    :param _ny: number of transmission data points in the vertical direction
    :param _dens: approximate density of objects [number of objects per mm^2]
    :param _edge_frac edge fraction defining how close to the edge can an object be placed on the object plane
    :param _sim_step_size: step size [m] that the simulation will take before trying to place an object (not used in some cases?)
    :param _obj_type: shape of objects to be generated: 1= rectangle, 2= ellipse, 3= triangle, 4= polygon, 5= random shapes
    :param _r_min_bw_obj:  minimum distance [m] between the centers of any two randomly placed objects 
    :param _obj_size_min: minimum allowed size of objects [m]
    :param _obj_size_max: maximum allowed size of objects [m]
    :param _size_dist: type of distribution of object sizes: 1- uniform, 2- normal (Gaussian), 3- Flory–Schulz
    :param _ang_min: minimum rotation angle of objects [degrees]
    :param _ang_max: maximum rotation angle of shape [degrees]
    :param _ang_dist: distribution of rotation angle. Choices are: 1=uniform, 2=normal(Gaussian), 3=Flory–Schulz
    :param _rand_alg: randomization algorithm:
        1=  uniform seeding (default); algorithm will create a set of objects positioned on a uniform rectangular grid,
            based on density, minimum and maximum object size, and the minimum distance between objects;
            it then applies random seeded noise to the point locations;
        2=  2D random walk: each object will be positioned using a 2D random walk algorithm;
    :param _obj_par1: optional parameter defining object shapes; its meaning depends on the choice of object type (i.e. the value of _obj_type):
        if _obj_type == 1 (rectangle): it is ratio of width to length (length is the longest dimension);
            leaving it as None (default) will set the width to length ratio to 1, unless _obj_par2 = True
            is selected to randomize the width to length ratio;
        if _obj_type == 2 (ellipse): ratio of minor to major semi-axes;
            leaving value as None will set this ratio to 1, unless _obj_par2 = True
            is selected to randomize the minor to major semi-axes ratio;
        if _obj_type == 3 (triangle): ratio of height (y, or "opposite") to width (x, or adjunct);
            leaving it as None will set height to width ratio to 1 unless _obj_par2 = True
            is selected to randomize the height to width ratio (default is an equilateral triangle);
        if _obj_type == 4 (regular polygon): number of polygon vertices;
            leaving it as None will give a hexagon (6 sides) unless _obj_par2 = True is selected
            to randomize the numer of polygon vertices (max. 12);
        if _obj_type == 5 (random shapes): list of identifiers of shapes to randomly generate
            (1= square, 2= circle, 3= triangle, 4= hexagon, 5= any of 1-4);
            leaving it as None will generate all shape options as [1,2,3,4,5]
    :param _obj_par2: optional switch (True / False) defining randomization of object shapes, depending on object type (i.e. the value of _obj_type):
        if _obj_type == 1 (rectangle): randomizes the width to length ratio;
        if _obj_type == 2 (ellipse): randomizes the minor to major semi-axes ratio;
        if _obj_type == 3 (triangle): randomizes the height (y, or "opposite") to width (x, or adjunct) ratio;
        if _obj_type == 4 (regular polygon): randomizes the numer of polygon vertices;
    :param _extTr: transmission outside the grid/mesh is zero (0), or it is same as on boundary (1)
    :param _e_start: initial value of photon energy (active if > 0 and if _delta and _atten_len are arrays / lists)
    :param _e_fin: final value of photon energy (active if > 0 and if _delta and _atten_len are arrays / lists)
    :param _ret: type of return: 'srw' returns SRWLOptT type transmission object which simulates the Sample (default),
        'img' returns processed PIL image object, 'all' returns both SRWLOptT object and PIL image.
    :return: by default, transmission (SRWLOptT) type optical element which simulates the Sample; other options are defined by _ret variable
    """

    ### All input parameters to include in return object
    input_parms = { #what for?
        "type": "sample_rnd_obj2d",
        "thickness": _thickness,
        "refractiveIndex": _delta,
        "attenuationLength": _atten_len,
        "horizontalRange": _rx,
        "verticalRange": _ry,
        "horizontalCenterCoordinate": _xc,
        "verticalCenterCoordinate": _yc,
        "horizontalPoints": _nx,
        "verticalPoints": _ny,
        "density": _dens,
        "edgeFraction": _edge_frac,
        "simulationStepSize": _sim_step_size,
        "objectType": _obj_type,
        "minDistanceBetweenObjects": _r_min_bw_obj, 
        "objectMinSize": _obj_size_min,
        "objectMaxSize": _obj_size_max,
        "sizeDistributionType": _size_dist, 
        "minRotationAngle": _ang_min,
        "maxRotationAngle": _ang_max,
        "angleDistributionType": _ang_dist,
        "randomizationAlgorithm": _rand_alg,
        "objectParameter1": _obj_par1,
        "objectParameter2": _obj_par2,
        "externalTransmissionType": _ext_tr,
        #"pointNumberPhotonEnergy": _ne, 
        "initialPhotonEnergy": _e_start,
        "finalPhotonPnergy": _e_fin,}

    import_err_msg = '"{}" library cannot be imported. Please install it first with "pip install {}".'

    ### Load package Numpy
    try:
        import numpy as np
    except:
        raise ImportError(import_err_msg.format('NumPy', 'numpy')) #OC24052020
        #print('NumPy can not be loaded. You may need to install numpy. If you are using pip, you can use the following ' + 
        #      'command to install it: \npip install numpy')

    ### Load package srwl_uti_smp_rand_obj2d.py
    try:
        #sys.path.append(os.path.join(os.path.dirname(__file__), '..')) #OC24052020 (commented-out)
        from srwl_uti_smp_rnd_obj2d import get_rnd_2D, on_pxy, get_r1j, uni_rnd_seed
    except:
        raise Exception('srwl_uti_smp_rnd_obj2d module failed to load. Please make sure that skimage module is installed. It can be installed with "pip install scikit-image" ') #OC24052020
        #print('srwl_uti_smp_rnd_obj2d.py functions get_rnd_2D, on_pxy, get_r1j, and uni_rnd_seed failed to load.')

    ### Begin time record of creating and placing objects
    #t0 = time.time();

    ### Test User Input Parameters
    if _obj_size_min > _obj_size_max:
        raise Exception('Minimum object size (_obj_size_min) is greater than maximum object size (_obj_size_max).') #OC24052020
        #print('Minimum object size (_obj_size_min) is greater than maximum object size (_obj_size_max).')
    if _ang_min > _ang_max:
        raise Exception('Minimum rotation angle (_ang_min) is greater than maximum rotation angle (_ang_max).') #OC24052020  
        #print('Minimum rotation angle (_ang_min) is greater than maximum rotation angle (_ang_max).' )    

    ### Generate random 2D disk with random uniform positions (that will be random-walk moved later)
      #Initilize the object center positions
    rx_pixels = _nx #:param rx_pixels = _nx: (srwl_uti_smp_rand_obj2d.py) box size (x), namely each frame size [array size (pixels)]
    ry_pixels = _ny #:param ry_pixels = _ny: (srwl_uti_smp_rand_obj2d.py) box size (y), namely each frame size [array size (pixels)]
    num = int(_dens * ((_rx*1000)*(_ry*1000))) #:param num: (srwl_uti_smp_rand_obj2d.py) particle number.

    ### Generate random 2D disk (get_rnd_2D)
    if _rand_alg == 2: # 2D random walk
      #Simulate a twoD random walk 
      #Each of the 2D points, for each step, move with random direction and with amplitue as amp 
        #:param px/py: numpy.meshgrid, coordination of x/y, shape will be [point number, point number]
        #:param amp: float, each step move amptitde
        #:param rmin: the min distance between any two points
        #:param try_max: try number to acheive rmin, if can't achevie print that point
        #:return: new px/py
        px = np.random.uniform( _edge_frac, 1-_edge_frac, num ) * rx_pixels #numpy.random.uniform(low, high, size)
        py = np.random.uniform( _edge_frac, 1-_edge_frac, num ) * ry_pixels #numpy.random.uniform(low, high, size) 
        amp = (_sim_step_size * px.shape[0])/_rx #convert simulation_step_size from units [m] to [# of cells]
        rmin = (_r_min_bw_obj * px.shape[0])/_rx #convert minimum_obj_dist from units [m] to [# of cells]
        #try:
        px, py = get_rnd_2D( px, py, amp = amp, rmin = rmin, try_max=10000)
        #print("srwl_uti_smp_rnd_obj2d returned a numpy meshgrid of " + str(num) + " objects ... ") #OC24052020 (commented-out)
        #except:
        #    print("srwl_uti_smp_rnd_obj2d failed to return a numpy meshgrid")
    else: #_rand_alg == 1 # random ~uniform seeding (default)
      #Random ~uniform seeding (_rand_alg = 1 #default): Algorithm will 
      # create a uniform set of objects based on: density, minimum and 
      # maximum object size, and the minimum distance between objects. 
      # It then applies random seeded noise to the point locations.
        rmin = (_r_min_bw_obj * num)/_rx #convert minimum_obj_dist from units [m] to [# of cells]
        size_max = int(_obj_size_max * (((rx_pixels/(_rx*2))+(rx_pixels/(_rx*2)))/2))
        #try:
        px, py = uni_rnd_seed(num, rx_pixels, ry_pixels, obj_max_size=size_max, min_dist=rmin) 
        #print("srwl_uti_smp_rnd_obj2d returned a numpy meshgrid of " + str(num) + " objects ... ")
        #except:
        #    print("srwl_uti_smp_rnd_obj2d failed to return a numpy meshgrid")

    ### Get radius of particles in units of pixels (to be used with srwl_uti_smp_rand_obj2d.py def on_pxy)
      #particle radius: size[pixels] = particle_radius [m] * (((rx_pixels/(_rx [m] *2))+(rx_pixels/(_rx [m] *2)))/2)
    size_max = int(_obj_size_max * (((rx_pixels/(_rx*2))+(rx_pixels/(_rx*2)))/2))
    size_min = int(_obj_size_min * (((rx_pixels/(_rx*2))+(rx_pixels/(_rx*2)))/2))
    if (size_min < 1):
        raise Exception('Particle radius too small. Try using a larger particle minimum size.') #OC24052020  
        #print("\nParticle radius too small for \"srwl_uti_smp_rnd_obj2d.py\" function \"on_pxy\" to map onto disk." + 
        #      " Try using a larger particle minimum size.\n")

    ### Place the shape in position (on_pxy). px, py is numpy.meshgrid, shape will be [N,N]
    #try:
    object_plane = on_pxy(px, py, bx = rx_pixels, by = ry_pixels,
                          _obj_type = _obj_type, _obj_size_min = size_min, _obj_size_max = size_max, _size_dist = _size_dist,
                          _ang_min = _ang_min, _ang_max = _ang_max, _ang_dist = _ang_dist,
                          _obj_par1 = _obj_par1, _obj_par2 = _obj_par2)
    #print('completed (lasted', round(time.time() - t0, 6), 's)')
    #except:
    #print("\"srwl_uti_smp_rnd_obj2d.py\" function \"on_pxy\" failed to return a numpy array. " +
    #      "You may need to install numpy and scikit-image. If you are using pip, you can use the following command to install them: " + 
    #      "\npip install numpy scikit-image")

    ### Select output type: srw transmission object, png, jpg, or tiff
    ### Get data from processed array
    limit_value = 255
    # processed data
    data = ((np.subtract(object_plane, object_plane.min(), dtype=np.float32)) / 
            (np.subtract(object_plane.max(), object_plane.min(), dtype=np.float32)) * limit_value)

#OC24052020 (commented-out)
##    #Save plot as png
##    #Load package matplotlib
##    if _ret == 'png':
##        try:
##            import matplotlib as mpl
##        except:
##            print('Matplotlib can not be loaded. You may need to install matplotlib. If you are using pip, you can use the following ' + 
##              "command to install it: \npip install matplotlib")
##        filename = os.path.join(os.getcwd(), _file_path, _file_name)
##        return mpl.image.imsave(filename + '.png', data, cmap = 'binary')
##    
##    #Save plot as jpg
##    #Load package matplotlib
##    elif _ret == 'jpg':
##        try:
##            import matplotlib as mpl
##        except:
##            print('Matplotlib can not be loaded. You may need to install matplotlib. If you are using pip, you can use the following ' + 
##              "command to install it: \npip install matplotlib")
##        filename = os.path.join(os.getcwd(), _file_path, _file_name)
##        return mpl.image.imsave(filename + '.jpg', data, cmap = 'binary')
##    
##    #Save plot as tiff
##    #Load package matplotlib
##    elif _ret == 'tiff' or _ret == 'tif':
##        try:
##            import matplotlib
##            matplotlib.use('Agg')
##            import matplotlib.pyplot as plt
##        except:
##            print('Matplotlib can not be loaded. You may need to install matplotlib. If you are using pip, you can use the following ' + 
##              "command to install it: \npip install matplotlib")
##        filename = os.path.join(os.getcwd(), _file_path, _file_name)
##
##        dpi = 200 # Arbitrary. The number of pixels in the image will always be identical
##        tif_height, tif_width = np.array(data.shape, dtype=float) / dpi
##        fig = plt.figure(figsize=(tif_width, tif_height), dpi=dpi)
##        ax = fig.add_axes([0, 0, 1, 1])
##        ax.axis('off')
##        ax.imshow(data, interpolation='none', cmap = 'binary')
##        
##        return fig.savefig(filename + '.tif', dpi=dpi, cmap = 'binary')
##    
##    else: #_ret == 'srw'

    #OC24052020
    img = None
    if((_ret == 'img') or (_ret == 'all')): #Return of PIL image object is required

        try:
            from PIL import Image
        except ImportError:
            raise ImportError(import_err_msg.format('PIL', 'pillow'))

        img = Image.fromarray(data.astype(np.uint8))
        if(_ret == 'img'): return img
 
    #elif(_ret == 'srw' or _ret == 'all'): #OC28052020
    #elif _ret == 'srw':
    ### Create _void_cen_rad: flat array/list of void center coordinates and radii: [x1, y1, r1, x2, y2, r2,...]
    #OC24052020 (commented-out)
    #_void_cen_rad_X = np.ravel(px)
    #_void_cen_rad_Y = np.ravel(py)
    #_void_cen_rad_R = np.ones(len(_void_cen_rad_X)) * ((_obj_size_max + _obj_size_min)/2)
    #_void_cen_rad = [None]*(len(_void_cen_rad_X)+len(_void_cen_rad_Y)+len(_void_cen_rad_R))
    #_void_cen_rad[::3] = _void_cen_rad_X # add X coordinates (starting at postion 0 with step size 3)
    #_void_cen_rad[1::3] = _void_cen_rad_Y # add Y coordinates (starting at postion 1 with step size 3)
    #_void_cen_rad[2::3] = _void_cen_rad_R # add radius size (starting at postion 2 with step size 3)

    # spectral refractive index decrement and/or attenuation length data
    ne = 1 #OC01062020
    #specPropAreDef = False # is spectral propagation defined?
    #if(_ne > 1): # if number of transmission data points vs photon energy > 1

    deltaIsArray = isinstance(_delta, list) or isinstance(_delta, array) or isinstance(_delta, np.ndarray) #OC01062020
    attenLenIsArray = isinstance(_atten_len, list) or isinstance(_atten_len, array) or isinstance(_atten_len, np.ndarray)
    #if((isinstance(_delta, list) or isinstance(_delta, array) or isinstance(_delta, np.ndarray)) and 
    #   (isinstance(_atten_len, list) or isinstance(_atten_len, array) or isinstance(_atten_len, np.ndarray))):
        #    lenDelta = len(_delta) # length refractive index list
        #    if((lenDelta == len(_atten_len)) and (lenDelta == _ne)): specPropAreDef = True
        #else: raise Exception("Inconsistent spectral refractive index decrement and/or attenuation length data")
    if deltaIsArray and attenLenIsArray: #OC01062020
        ne = len(_delta)
        ne1 = len(_atten_len)
        if(ne > ne1): ne = ne1
    else:
        if deltaIsArray:
            ne = len(_delta)
            arAttenLen = [_atten_len]*ne
            _atten_len = arAttenLen
        elif attenLenIsArray:
            ne = len(_atten_len)
            arDelta = [_delta]*ne
            _delta = arDelta

    # process numpy array using physical properties such as thickness/attenuation-length/etc.
    thickByLim = _thickness/limit_value # thickness by limit
    nxny = _nx*_ny # transmission data points horizontal x vertical 
    miHalfThickByLimByAttenLen = None # initalize half thick limit by attenuation length
    miThickDeltaByLim = None # initalize thick refractive index by limit

    #creating arTr: (srwlib.SRWLOptT) complex C-aligned data array (of 2*ne*nx*ny length) storing amplitude 
    if(ne <= 1): #OC01062020 # if number of transmission data points vs photon energy <= 1
    #if(_ne <= 1): # if number of transmission data points vs photon energy <= 1
        miHalfThickByLimByAttenLen = -0.5*thickByLim/_atten_len # (-1/2)*(thickness/limit_value)/(attenuation length)
        miThickDeltaByLim = -thickByLim*_delta # -(thickness/limit_value)*(refractive index)

        # amplitude transmission data
        amplTransm = np.exp(miHalfThickByLimByAttenLen*data)
        # optical path difference
        optPathDiff =  miThickDeltaByLim*data 

        #Create arTr: (srwlib.SRWLOptT) complex C-aligned data array (of 2*ne*nx*ny length) storing amplitude 
        # transmission and optical path difference as function of transverse coordinates
        arTr = np.empty((2*nxny), dtype=float) 
        arTr[0::2] = np.reshape(amplTransm, nxny)
        arTr[1::2] = np.reshape(optPathDiff, nxny)

    else: # if number of transmission data points vs photon energy > 1
        #two_ne = 2*_ne
        two_ne = 2*ne #OC01062020
        arTr = np.empty((nxny*two_ne), dtype=float)

        #if(specPropAreDef == False):
        #    miHalfThickByLimByAttenLen = -0.5*thickByLim/_atten_len# (-1/2)*(thickness/limit_value)/(attenuation length)
        #    miThickDeltaByLim = -thickByLim*_delta# -(thickness/limit_value)*(refractive index)

        #for ie in range(_ne):
        for ie in range(ne): #OC01062020
            #if(specPropAreDef):
            #OC01062020
            miHalfThickByLimByAttenLen = -0.5*thickByLim/_atten_len[ie] # (-1/2)*(thickness/limit_value)/(attenuation length)[ie]
            miThickDeltaByLim = -thickByLim*_delta[ie] # -(thickness/limit_value)*(refractive index)[ie]

            # amplitude transmission data
            amplTransm = np.exp(miHalfThickByLimByAttenLen*data)
            # optical path difference
            optPathDiff = miThickDeltaByLim*data

            #Create arTr: (srwlib.SRWLOptT) complex C-aligned data array (of 2*ne*nx*ny length) storing amplitude transmission and optical path difference as function of transverse position
            # transmission and optical path difference as function of transverse coordinates
            two_ie = 2*ie
            arTr[two_ie::two_ne] = np.reshape(amplTransm, nxny)
            arTr[(two_ie + 1)::two_ne] = np.reshape(optPathDiff, nxny)

    ### Create transmission (SRWLOptT) type optical element
    #opT = srwlib.SRWLOptT(_nx=_nx, _ny=_ny, _rx=_rx, _ry=_ry, _arTr=arTr, _extTr=_ext_tr, _x=_xc, _y=_yc, _ne=_ne, _eStart=_e_start, _eFin=_e_fin)
    opT = srwlib.SRWLOptT(_nx=_nx, _ny=_ny, _rx=_rx, _ry=_ry, _arTr=arTr, _extTr=_ext_tr, _x=_xc, _y=_yc, _ne=ne, _eStart=_e_start, _eFin=_e_fin) #OC01062020
    opT.input_parms = input_parms # add user input parameters to transmission optical element

    #OC28052020
    if(_ret == 'srw'): return opT 
    elif(_ret == 'all'): return opT, img
    else: return None



# -*- coding: utf-8 -*-
#############################################################################
# Generate a 2D random pattern of disks
# Authors/Contributors: Yugang Zhang, Rebecca Ann Coles
# March 01, 2019
# April 11, 2020
#############################################################################

import os
import sys
import numpy as np

from skimage.draw import polygon, circle, ellipse

# ********************** Get the distance between point jj and the other points
def get_r1j( px, py, jj):
    '''Used with Yugang random walk function.
    Get the distance between point jj and the other points

    :param px: mesgrid, coordination of x, shape will be [point number, point number]
    :param py: mesgrid, coordination of y, shape will be [point number, point number]
    :param jj: the index of the position
    :return: the distance between the j and other points, shape be point number -1
    '''

    pxf = np.ravel( px )
    pyf = np.ravel( py )
    pxn = np.delete( pxf, jj, 0)
    pyn = np.delete( pyf, jj, 0)
    dd = np.sqrt(  ( pxn - pxf[jj] )**2 + ( pyn - pyf[jj] )**2 )
    return dd


# ********************** Check the distance between point jj and other points 
def chk_r1j_rmin(px,py, jj, rmin):
    '''Used with Yugang random walk function.
    Check the distance between point j and other points if all the distances 
    larger than rmin, return True.

    :param px: mesgrid, coordination of x, shape will be [point number, point number]
    :param py: mesgrid, coordination of y, shape will be [point number, point number]
    :param jj: the index of the position
    :param rmin: the object minimum distance from other objects.
    :return: bool, True = distance is smaller than rmin, need to move position.
        False = distance larger than rmin, no need to move position.
    '''

    r1j = get_r1j( px, py, jj)

    if len(np.where( r1j < rmin)[0]) >= 1:  #distance smaller than rmin, need to move position 
        return True
    else:
        return False


# ********************** Move 2D points with random direction and with amplitue as amp for each step
def mv_2D( px, py, amp): 
    '''Used with Yugang random walk function.
    Move 2D points with random direction and with amplitue as amp for each step 
    
    :param px: mesgrid, coordination of x, shape will be [point number, point number]
    :param py: mesgrid, coordination of y, shape will be [point number, point number]
    :param amp: amplitue of each step.
    :return: px/py mesgrid, coordination of x/y, shape will be [point number, point number]
        or px/py: one float, in case of move one point
    '''

    NA=False
    if not isinstance(px, (list, np.ndarray) ):
        px = np.array( [px])
        py = np.array( [py] )
        NA = True

    theta = np.random.uniform(-0.5,0.5,(len(px),len(py))) * 2* np.pi
    dx = amp * np.cos( theta )
    dy = amp * np.sin( theta ) 

    if NA:
        return (px+dx)[0], (py+dy)[0]
    else:
        return px + dx, py + dy


# ********************** Simulate a 2D random walk 
def get_rnd_2D( px, py, amp, rmin=16, try_max=1000):
    '''Yugang's random walk function.
    Simulate a 2D random walk. Each of the 2D points, for each step, move with 
    random direction and with amplitue as amp.

    :param px: mesgrid, coordination of x, shape will be [point number, point number]
    :param py: mesgrid, coordination of y, shape will be [point number, point number]
    :param amp: amplitue of each step.
    :param rmin: the object minimum distance from other objects.
    :param try_max: try number to acheive rmin, if can't achevie print that point
    :return: px/py mesgrid, coordination of x/y, shape will be [point number, point number]
    '''

    px = np.ravel(px)
    py = np.ravel(py)
    for ii in range( px.size):
        trys = 0
        mov = True
        while mov:
            px[ii], py[ii] =  mv_2D( px[ii], py[ii], amp)
            mov = chk_r1j_rmin(px,py, ii, rmin)
            trys +=1
            if trys > try_max:
                mov = False
                print(ii, 'Not satify the rmin distance')
    return px,py


def uni_rnd_seed(num, rx_pixels, ry_pixels, obj_max_size, min_dist=16): #RAC25032020
    '''
    Uniform Random Seed Object Locations
    Evenly space "num" points over a (rx, ry) grid.

    :param num: particle number.
    :param rx_pixels: range of the horizontal coordinate [pixels] for which 
        the transmission is defined.
    :param ry_pixels: range of the vertical coordinate [pixels] for which 
        the transmission is defined.
    :param _obj_size_max: maximum allowed size of objects [pixels]
    :param min_dist = 16: minimum distance [pixels] between objects.    

    For num_x points in x-direction with spacing delta_x in a 2D array of width
    "w" and height "h":
        num_x = sqrt[(w/h)*num + (w-h)^2 / (4*h^2)] - [(w-h) / (2*h)]
        
    For num_y points in y-direction with spacing delta_y in a 2D array of width
    "w" and height "h":
        num_y = num / num_x

    To get (x,y) point pairs you need delta_x = delta_y. If delta_x != delta_y
    for num objects, num is reduced until delta_x = delta_y.
    '''
    # Get spacing (delta) between points. 
    #  If spacing can't be compleated with the requested
    #  number of points, the density of points will be reduced until 
    #  the points can all be uniformly spaced.
    final_num_points = num #initalize to maximum number of points
    delta_x = 0 #initalize to delta_x != delta_y
    delta_y = 1 #initalize to delta_x != delta_y
    dist_edge = obj_max_size*2 #minimum distance of an object from the edge of the surface
    rx_boundry = rx_pixels-dist_edge*2 #end edge boundry for objects
    ry_boundry = ry_pixels-dist_edge*2 #top edge boundry for objects

    while delta_x != delta_y: #delta_x=delta_y needed to evenly spread objects
        num_x = int(np.sqrt( (rx_boundry/ry_boundry)*final_num_points + 
                            np.power((rx_boundry-ry_boundry),2) / 
                            np.power((4*ry_boundry),2) ) - 
                            (rx_boundry-ry_boundry)/(2*ry_boundry))
        num_y = int(final_num_points/num_x)
        delta_x = int(rx_boundry/(num_x-1))
        delta_y = int(ry_boundry/(num_y-1))
        if delta_x != delta_y:
            # return a default set of objects if delta_x=delta_y fails
            if final_num_points <= 0:
                delta_x = delta_y = 1
                final_num_points = final_num_points/2
            # reduce num until delta_x=delta_y
            final_num_points = final_num_points-1
        else:
            # if delta_x=delta_y, set delta
            delta = delta_x

    # Construct px and py made up of uniformly spaced points in (rx, ry)
    px_edge = [] #x coord list of uniformly spaced points
    py_edge = [] #y coord list of uniformly spaced points
    rows = num_y #number of rows (y values)
    columns = num_x #number of columns (x values) 
    for ii in range(rows-1):
        #fill px and py with uniformly spaced points (row by row)
        row_value = ii*delta
        for jj in range(columns-1):
            px_edge.append(jj*delta)
            py_edge.append(row_value)

    # Enforce object distance from left an bottom surface boundry        
    px = [kk+dist_edge for kk in px_edge]
    py = [kk+dist_edge for kk in py_edge]

    # Create random noise to perturb points
    minimum_movement = obj_max_size + min_dist
    if delta <= minimum_movement:
        print("Warning: object_maximum_size + minimum_dist_between_obj is" 
              " greater than perfectly uniform object spaceing. To preserve"
              + " minimum_dist_between_obj you will need to reduce object_maximum_size"
              + " or decrease object density.")
    noise = np.random.uniform(low=minimum_movement,
                              high=delta,
                              size=(len(px), 2))

    #add noise to px and py
    px = np.asarray(px, dtype=np.float32)
    py = np.asarray(py, dtype=np.float32) 
    px += noise[:,0]
    py += noise[:,1]

    return py, px

# ********************** Place shape on meshgrid
def on_pxy(px, py, bx=100, by=None,
           _obj_type=1, _obj_size_min = 10, _obj_size_max = 10, _size_dist = 1,
           _ang_min = 0, _ang_max = 0, _ang_dist = 1,
           _obj_par1 = None, _obj_par2 = None): #RAC25032020
    '''Place shapes on (px,py) points

    :param px: mesgrid, coordination of x, shape will be [point number, point number]
    :param py: mesgrid, coordination of y, shape will be [point number, point number]
    :param bx: box size (x), namely each frame size [array size (pixels)]
    :param by: box size (y), namely each frame size [array size (pixels)]
    :param _obj_type = 1: shape of objects. 
        Choices: 1=rectangle, 2=ellipse, 3=triangle, 4=polygon, 5=random_shapes
    :param _obj_size_min = 10: minimum allowed size of objects [pixels]
    :param _obj_size_max = 10: maximum allowed size of objects [pixels]
    :param _size_dist = 1: distribution of sizes. 
        Choices are: 1=uniform, 2=normal(Gaussian), 3=Flory–Schulz
    :param _ang_min = 0: minimum rotation angle of shape [degrees].
    :param _ang_max = 0: maximum rotation angle of shape [degrees].
    :param _ang_dist = 1: distribution of rotation angle. 
        Choices are: 1=uniform, 2=normal(Gaussian), 3=Flory–Schulz
    :param _obj_par1 = None: 
        rectangle: ratio of width to length (length is the longest dimension). 
            Leaving value as None will give 1.0 (square) as width to length 
            ratio unless _obj_par2 = True is selected to randomize the 
            width to length ratio.
        ellipse: ratio of minor to major semi-axes. 
            Leaving value as None will give 1.0 (circle) as minor to major 
            semi-axes ratio unless _obj_par2 = True is selected to 
            randomize the minor to major semi-axes ratio.
        triangle: ratio of height (y, or "opposite") to width (x, or adjunct). 
            Leaving value as None will give 1.0 (1/1) as height to width 
            ratio unless _obj_par2 = True is selected to randomize the 
            height to width ratio (default is an equilateral triangle).
        regular polygon: number of polygon vertices.
            Leaving value as None will give a hexagon (6 sides)
            unless _obj_par2 = True is selected to randomize the numer of 
            polygon vertices. Max vertices = 12.
        random shapes: Which shapes to randomly generate.
            Choices are:
                [1='rectangle',2='ellipse',3='triangle',4='polygon'].
            Leaving value as None will give all shape options as:[1,2,3,4]
    :param _obj_par2 = None:
        rectangle: value set to True will randomize the width to length ratio.
        ellipse: Value set to True will randomize the minor to major semi-axes ratio.
        triangle: value set to True will randomize the height (y, or "opposite")
            to width (x, or adjunct) ratio.
        regular polygon: Value set to True will randomize the numer of polygon 
            vertices.
    :return:array with shapes at (px,py) locations.
    '''

    if by is None:
        by = bx
    dd = np.zeros( [bx,by] , dtype = bool)   

    for i in range(px.size): 
        cx = np.ravel(px)[i]
        cy = np.ravel(py)[i]
        xx, yy = get_shape(cx, cy, _obj_type = _obj_type, 
                           _obj_size_min = _obj_size_min, _obj_size_max = _obj_size_max, _size_dist = _size_dist,
                           _ang_min = _ang_min, _ang_max = _ang_max, _ang_dist = _ang_dist,
                           _obj_par1 = _obj_par1, _obj_par2 = _obj_par2)
        ww = np.where( ( (xx>=0)  & (xx<bx) & (yy>=0) & (yy<by) ) )            
        dd[xx[ww],yy[ww]]=1 

    return dd


# ********************** Get shape in accordance with user specified parameters.
def get_shape(cx, cy, _obj_type = 1, 
              _obj_size_min = 10, _obj_size_max = 10, _size_dist = 1,
              _ang_min = 0, _ang_max = 0, _ang_dist = 1,
              _obj_par1 = None, _obj_par2 = None): #RAC25032020
    '''Get shape in accordance with user specified parameters.

    :param cx: x coordinate of object center.
    :param cy: y coordinate of object center.
    :param _obj_type = 1: shape of objects. Choices are: 
        1=rectangle, 2=ellipse, 3=triangle, 4=ploygon, 5=random_shapes
    :param _obj_size_min = 10: minimum allowed size of objects [pixels]
    :param _obj_size_max = 10: maximum allowed size of objects [pixels]
    :param _size_dist = 1: distribution of sizes. Choices are: 
        1=uniform, 2=normal(Gaussian), 3=Flory–Schulz
    :param _ang_min = 0: minimum rotation angle of objects [degrees].
    :param _ang_max = 0: maximum rotation angle of objects [degrees].
    :param _ang_dist = 1: distribution of rotation angle. Choices are: 
        1=uniform, 2=normal(Gaussian), 3=Flory–Schulz
    :param _obj_par1 = None: 
        rectangle: ratio of width to length (length is the longest dimension). 
            Leaving value as None will give 1.0 (square) as width to length 
            ratio unless _obj_par2 = True is selected to randomize the 
            width to length ratio.
        ellipse: ratio of minor to major semi-axes. 
            Leaving value as None will give 1.0 (circle) as minor to major 
            semi-axes ratio unless _obj_par2 = True is selected to 
            randomize the minor to major semi-axes ratio.
        triangle: ratio of height (y, or "opposite") to width (x, or adjunct). 
            Leaving value as None will give 1.0 (1/1) as height to width 
            ratio unless _obj_par2 = True is selected to randomize the 
            height to width ratio (default is an equilateral triangle).
        regular polygon: number of polygon vertices.
            Leaving value as None will give a hexagon (6 sides)
            unless _obj_par2 = True is selected to randomize the numer of 
            polygon vertices.
        random shapes: Which shapes to randomly generate.
            Choices are [1='rectangle',2='ellipse',3='triangle',4='polygon'].
            Leaving value as None will give all shape options as:
                [1,2,3,4]
    :param _obj_par2 = None:
        rectangle: value set to True will randomize the width 
            to length ratio.
        ellipse: Value set to True will randomize the minor to major 
            semi-axes ratio.
        triangle: value set to True will randomize the height (y, or "opposite")
            to width (x, or adjunct) ratio.
        regular polygon: Value set to True will randomize the numer of 
            polygon vertices. Max vertices = 12.
    '''

    if _obj_type == 1: #rectangle
        return get_rec(cx, cy, 
                       _obj_size_min = _obj_size_min, 
                       _obj_size_max = _obj_size_max, 
                       _size_dist = _size_dist, 
                       _ang_min = _ang_min, 
                       _ang_max = _ang_max, 
                       _ang_dist = _ang_dist, 
                       _obj_par1 = _obj_par1, 
                       _obj_par2 = _obj_par2)   
    elif _obj_type == 2: #ellipse
        return get_elp(cx, cy, 
                       _obj_size_min = _obj_size_min, 
                       _obj_size_max = _obj_size_max, 
                       _size_dist = _size_dist,
                       _ang_min = _ang_min, 
                       _ang_max = _ang_max, 
                       _ang_dist = _ang_dist, 
                       _obj_par1 = _obj_par1, 
                       _obj_par2 = _obj_par2)  
    elif _obj_type == 3: #triangle
        return get_tri(cx, cy, 
                       _obj_size_min = _obj_size_min, 
                       _obj_size_max = _obj_size_max, 
                       _size_dist = _size_dist,
                       _ang_min = _ang_min, 
                       _ang_max = _ang_max, 
                       _ang_dist = _ang_dist, 
                       _obj_par1 = _obj_par1, 
                       _obj_par2 = _obj_par2)
    elif _obj_type == 4: #polygon
        return get_pol(cx, cy, 
                       _obj_size_min = _obj_size_min, 
                       _obj_size_max = _obj_size_max, 
                       _size_dist = _size_dist,
                       _ang_min = _ang_min, 
                       _ang_max = _ang_max, 
                       _ang_dist = _ang_dist, 
                       _obj_par1 = _obj_par1, 
                       _obj_par2 = _obj_par2) 
    elif _obj_type == 5: #random
        return get_rnd(cx, cy, 
                       _obj_size_min = _obj_size_min, 
                       _obj_size_max = _obj_size_max, 
                       _size_dist = _size_dist, 
                       _ang_min = _ang_min, 
                       _ang_max = _ang_max, 
                       _ang_dist = _ang_dist, 
                       _obj_par1 = _obj_par1)
    else: #ellipse
        return get_elp(cx, cy, 
                       _obj_size_min = _obj_size_min, 
                       _obj_size_max = _obj_size_max, 
                       _size_dist = _size_dist,
                       _ang_min = _ang_min, 
                       _ang_max = _ang_max, 
                       _ang_dist = _ang_dist, 
                       _obj_par1 = _obj_par1, 
                       _obj_par2 = _obj_par2)

# ********************** Create rectangle
def get_rec(cx, cy, _obj_size_min = 10, _obj_size_max = 10, _size_dist = 1,
            _ang_min = 0, _ang_max = 0, _ang_dist = 1,
            _obj_par1 = None, _obj_par2 = None): #RAC25032020
    '''Rectangles

    :param cx: x coordinate of rectangle center.
    :param cy: y coordinate of rectangle center.
    :param _obj_size_min = 10: minimum allowed size of rectangle [pixels]
    :param _obj_size_max = 10: maximum allowed size of rectangle [pixels]
    :param _size_dist = 1: distribution of sizes. Choices are: 
        1=uniform, 2=normal(Gaussian), 3=Flory–Schulz
    :param _ang_min = 0: minimum rotation angle of rectangle [degrees].
    :param _ang_max = 0: maximum rotation angle of rectangle [degrees].
    :param _ang_dist = 1: distribution of rotation angle. Choices are: 
        1=uniform, 2=normal(Gaussian), 3=Flory–Schulz
    :param _obj_par1 = None: ratio of width to length (length is the 
        longest dimension). Leaving value as None will give 1.0 (square) as 
        width to length ratio unless _obj_par2 = True is selected to 
        randomize the width to length ratio.
    :param _obj_par2 = None: value set to True will randomize the width 
        to length ratio.
    '''

    #Use _obj_par1 (width to length ratio) or create random ratio 
    # if _obj_par2 = True
    default_side_ratio = 1 # 1/1 Square
    width, length = obj_opt_par(_obj_type = 1,
                                _obj_size_min = _obj_size_min,
                                _obj_size_max = _obj_size_max,
                                _size_dist = _size_dist,
                                _default_axis_ratio = default_side_ratio,
                                _obj_par1 = _obj_par1,
                                _obj_par2 = _obj_par2)
    
    #Create rectangle and get vertices
    _rows = np.array([cy-(width), cy-(width), 
                      cy+(width), cy+(width)])
    _columns = np.array([cx-(length), cx+(length), 
                         cx+(length), cx-(length)])

    ##Rotate the coordinates of vertices of the rectangle
    # Rotate a point counterclockwise by a given angle (degrees) around center
    rotated_rows, rotated_columns = rot_obj_vert(cx, cy, _rows, _columns,
                                                 _ang_min = _ang_min, _ang_max = _ang_max,
                                                 _ang_dist = _ang_dist)

    #Create rectangle (will use poloygon function so that we can create the rotated object)
    return polygon(rotated_rows, rotated_columns) 


# ********************** Create ellipse    
def get_elp(cx, cy, _obj_size_min = 10, _obj_size_max = 10, _size_dist = 1,
            _ang_min = 0, _ang_max = 0, _ang_dist = 1,
            _obj_par1 = None, _obj_par2 = None): #RAC25032020
    '''Ellipses

    :param cx = x: coordinate of ellipse center.
    :param cy = y: coordinate of ellipse center.
    :param _obj_size_min = 10: minimum allowed size of ellipse [pixels]
    :param _obj_size_max = 10: maximum allowed size of ellipse [pixels]
    :param _size_dist = 1: distribution of sizes. Choices are: 
        1=uniform, 2=normal(Gaussian), 3=Flory–Schulz
    :param _ang_min = 0: minimum rotation angle of ellipse [degrees].
    :param _ang_max = 0: maximum rotation angle of ellipse [degrees].
    :param _ang_dist = 1: distribution of rotation angle. Choices are: 
        1=uniform, 2=normal(Gaussian), 3=Flory–Schulz
    :param _obj_par1 = None: ratio of minor to major semi-axes. Leaving 
        value as None will give 1.0 (circle) as minor to major semi-axes 
        ratio unless _obj_par2 = True is selected to randomize the ratio.
    :param _obj_par2 = None: value set to True will randomize the minor 
        to major semi-axes ratio.
    '''

    #Use _obj_par1 (minor to major semi-axes ratio) or create 
    #  random ratio if _obj_par2 = True
    default_axis_ratio = 1.0 #circle
    r_radius, c_radius = obj_opt_par(_obj_type = 2,
                                     _obj_size_min = _obj_size_min,
                                     _obj_size_max = _obj_size_max,
                                     _size_dist = _size_dist,
                                     _default_axis_ratio = default_axis_ratio,
                                     _obj_par1 = _obj_par1,
                                     _obj_par2 = _obj_par2) #minor axis, major axis

    try:
        _ang_max >= _ang_min
    except:
        print("Maximum angle of rotation is less than than minimum angle of rotation")

    #Select random rotation angle for ellipse
    _obj_ang = get_dist(_min_value = _ang_min, _max_value = _ang_max, _dist = _ang_dist)

    #Create ellipse
    return ellipse(cx, cy, r_radius, c_radius, rotation=np.deg2rad(-_obj_ang))


# ********************** Create triangle
def get_tri(cx, cy, _obj_size_min = 10, _obj_size_max = 10, _size_dist = 1,
            _ang_min = 0, _ang_max = 0, _ang_dist = 1, _obj_par1 = None, _obj_par2 = None): #RAC25032020
    '''Triangles

    :param cx: x coordinate of triangle center.
    :param cy: y coordinate of triangle center.
    :param _obj_size_min = 10: minimum allowed size of triangle [pixels]
    :param _obj_size_max = 10: maximum allowed size of triangle [pixels]
    :param _size_dist = 1: distribution of sizes. Choices are: 
        1=uniform, 2=normal(Gaussian), 3=Flory–Schulz
    :param _ang_min = 0: minimum rotation angle of triangle [degrees].
    :param _ang_max = 0: maximum rotation angle of triangle [degrees].
    :param _ang_dist = 1: distribution of rotation angle. Choices are: 
        1=uniform, 2=normal(Gaussian), 3=Flory–Schulz
    :param _obj_par1 = None: ratio of height (y, or "opposite") to width 
        (x, or adjunct). Leaving value as None will give 1.0 (1/1) as 
        height to width ratio unless _obj_par2 = True is selected 
        to randomize the height to width ratio (default is an 
        equilateral triangle).
    :param _obj_par2 = None: value set to True will randomize the height 
        to width ratio.
    '''

    #Use _obj_par1 (height (y, or "opposite") to width (x, or adjunct) 
    #  ratio) or create random ratio if _obj_par2 = True
    default_side_ratio = 1.0 #1/1 equilateral triangle
    height, width = obj_opt_par(_obj_type = 3,
                                _obj_size_min = _obj_size_min,
                                _obj_size_max = _obj_size_max,
                                _size_dist = _size_dist,
                                _default_axis_ratio = default_side_ratio,
                                _obj_par1 = _obj_par1,
                                _obj_par2 = _obj_par2)

    #Create triangle and get vertices
    _rows = np.array([cy-(height), cy+(height), cy+(height)])
    _columns = np.array([cx-(width), cx+(width), cx-(width)])

    ##Rotate the coordinates of vertices of the triangle
    # Rotate a point counterclockwise by a given angle (degrees) around center
    rotated_rows, rotated_columns = rot_obj_vert(cx, cy, _rows, _columns,
                                                 _ang_min = _ang_min,
                                                 _ang_max = _ang_max,
                                                 _ang_dist = _ang_dist)

    #Create triangle (will use poloygon function so that we can create the rotated the object)
    return polygon(rotated_rows, rotated_columns)    


# ********************** Create polygon
def get_pol(cx, cy, _obj_size_min = 10, _obj_size_max = 10, _size_dist = 1,
            _ang_min = 0, _ang_max = 0, _ang_dist = 1, _obj_par1 = None, _obj_par2 = None): #RAC25032020
    '''Regular Polygon

    :param cx: x coordinate of polygon center.
    :param cy: y coordinate of polygon center.
    :param _obj_size_min = 10: minimum allowed size of polygon [pixels].
    :param _obj_size_max = 10: maximum allowed size of polygon [pixels].
    :param _size_dist = 1: distribution of sizes. Choices are: 
        1=uniform, 2=normal(Gaussian), 3=Flory–Schulz
    :param _ang_min = 0: minimum rotation angle of ploygon [degrees].
    :param _ang_max = 0: maximum rotation angle of polygon [degrees].
    :param _ang_dist = 1: distribution of rotation angle. Choices are: 
        1=uniform, 2=normal(Gaussian), 3=Flory–Schulz
    :param _obj_par1 = None: number of polygon vertices.
        Leaving value as None will give a hexagon (6 sides) unless 
        _obj_par2 = True is selected to randomize the numer of polygon 
        vertices.
    :param _obj_par2 = None: Value set to True will randomize the numer 
        of polygon vertices. Max vertices = 12.
    '''

    #Use _obj_par1 (number of vertices) to get radius and num_polygon_vertices
    #  or create random number of vertices if _obj_par2 = True
    #  Function "obj_opt_par" sets the default number of regular polygon
    #  vertices to "num_polygon_vertices = 6" (hexagon).
    radius, num_polygon_vertices = obj_opt_par(_obj_type = 4,
                                               _obj_size_min = _obj_size_min,
                                               _obj_size_max = _obj_size_max,
                                               _size_dist = _size_dist,
                                               _obj_par1 = _obj_par1,
                                               _obj_par2 = _obj_par2)

    #Create regular polygon and get vertices
    # rr = radius, ii = vertex, nn = total vertices, aa = angle
    # vx = cx + rr * cos(2 * pi * ii / nn + aa)
    # vy = cy + rr * sin(2 * pi * ii / nn + aa)
    _columns = [] #x values
    _rows = [] #y values
    aa = 360/num_polygon_vertices
    for ii in range(num_polygon_vertices):
        _columns.append(int(cx + radius * np.cos(2 * np.pi * ii / num_polygon_vertices + aa)))
        _rows.append(int(cy + radius * np.sin(2 * np.pi * ii / num_polygon_vertices + aa)))

    ##Rotate the coordinates of vertices of the regular polygon
    # Rotate a point counterclockwise by a given angle (degrees) around center
    rotated_rows, rotated_columns = rot_obj_vert(cx, cy, 
                                                 _rows = np.array(_rows), 
                                                 _columns = np.array(_columns), 
                                                 _ang_min = _ang_min, 
                                                 _ang_max = _ang_max, 
                                                 _ang_dist = _ang_dist)

    #Create polygon
    return polygon(rotated_rows, rotated_columns)  


# ********************** Create random shape
def get_rnd(cx, cy, _obj_size_min = 10, _obj_size_max = 10, _size_dist = 1,
            _ang_min = 0, _ang_max = 0, _ang_dist = 1, _obj_par1 = None): #RAC25032020
    '''Random Shapes

    :param cx: x coordinate of shape center.
    :param cy: y coordinate of shape center.
    :param _obj_size_min = 10: minimum allowed size of shape [pixels].
    :param _obj_size_max = 10: maximum allowed size of shape [pixels].
    :param _size_dist = 1: distribution of sizes. Choices are: 
        1=uniform, 2=normal(Gaussian), 3=Flory–Schulz
    :param _ang_min = 0: minimum rotation angle of shape [degrees].
    :param _ang_max = 0: maximum rotation angle of shape [degrees].
    :param _ang_dist = 1: distribution of rotation angle. Choices are: 
        1=uniform, 2=normal(Gaussian), 3=Flory–Schulz
    :param _obj_par1 = None: Which shapes to randomly generate.
        Choices are [1='rectangle',2='ellipse',3='triangle',4='polygon'].
        Leaving value as None will give all shape options as: [1,2,3,4]
    '''

    #Use _obj_par1 to get random shape from list of shapes
    random_shape = obj_opt_par( _obj_type = 5, _obj_par1 = _obj_par1)

    #Generate shape
    if random_shape == 1: #rectangle
        return get_rec(cx, cy,
                       _obj_size_min = _obj_size_min,
                       _obj_size_max = _obj_size_max,
                       _size_dist = _size_dist,
                       _ang_min = _ang_min,
                       _ang_max = _ang_max,
                       _ang_dist = _ang_dist)
    elif random_shape == 2: #ellipse
        return get_elp(cx, cy,
                       _obj_size_min = _obj_size_min,
                       _obj_size_max = _obj_size_max,
                       _size_dist = _size_dist,
                       _ang_min = _ang_min,
                       _ang_max = _ang_max,
                       _ang_dist = _ang_dist)
    elif random_shape == 3: #triangle
        return get_tri(cx, cy,
                       _obj_size_min = _obj_size_min,
                       _obj_size_max = _obj_size_max,
                       _size_dist = _size_dist,
                       _ang_min = _ang_min,
                       _ang_max = _ang_max,
                       _ang_dist = _ang_dist)
    elif random_shape == 4: #polygon
        return get_pol(cx, cy,
                       _obj_size_min = _obj_size_min,
                       _obj_size_max = _obj_size_max,
                       _size_dist = _size_dist,
                       _ang_min = _ang_min,
                       _ang_max = _ang_max,
                       _ang_dist = _ang_dist)
    else: #ellipse
        return get_elp(cx, cy,
                       _obj_size_min = _obj_size_min,
                       _obj_size_max = _obj_size_max,
                       _size_dist = _size_dist,
                       _ang_min = _ang_min,
                       _ang_max = _ang_max,
                       _ang_dist = _ang_dist)


# ********************** Use optional object parameters to get optional shape parameters
def obj_opt_par(_obj_type = 1, _obj_size_min = 10, _obj_size_max = 10, _size_dist = 1,
                _default_axis_ratio = 1.0, _obj_par1 = None, _obj_par2 = None): #RAC25032020
    '''Object Optional Parameters

        :param _obj_type = 1: shape of objects. Choices are: 
            1=rectangle, 2=ellipse, 3=triangle, 4=polygon, 5=random_shapes
        :param _obj_size_min = 10: minimum allowed size of rectangle [pixels]
        :param _obj_size_max = 10: maximum allowed size of rectangle [pixels]
        :param _size_dist = 1: distribution of sizes. Choices are: 
            1=uniform, 2=normal(Gaussian), 3=Flory–Schulz
        :param _obj_par1 = None: 
            rectangle: ratio of width to length (length is the 
                longest dimension). 
                Leaving value as None will give 1.0 (square) as width to length 
                ratio unless _obj_par2 = True is selected to randomize the 
                width to length ratio.
            ellipse: ratio of minor to major semi-axes. 
                Leaving value as None will give 1.0 (circle) as minor to major 
                semi-axes ratio unless _obj_par2 = True is selected to 
                randomize the minor to major semi-axes ratio.
            triangle: ratio of height (y, or "opposite") to width 
                (x, or adjunct). Leaving value as None will give 1.0 (1/1) as 
                height to width ratio unless _obj_par2 = True is selected 
                to randomize the height to width ratio (default is an 
                equilateral triangle).
            regular polygon: number of sides for polygon.
                Leaving value as None will give a hexagon (6 sides)
                unless _obj_par2 = True is selected to randomize the numer of 
                polygon vertices.
            random shapes: Which shapes to randomly generate.
                Choices are [1='rectangle',2='ellipse',3='triangle',4='polygon'].
                Leaving value as None will give all shape options as:
                    [1,2,3,4]
        :param _obj_par2 = None:
            rectangle: value set to True will randomize the width 
                to length ratio.
            ellipse: Value set to True will randomize the minor to major 
                semi-axes ratio.
            triangle: None: value set to True will randomize the height 
                to width ratio.
            regular polygon: Value set to True will randomize the numer of polygon 
                vertices. Max vertices = 12.
    '''

    #Default number of regular polygon vertices
    num_polygon_vertices = 6 #default: hexagon (all 6 sides)
    min_number_sides = 3
    max_number_sides = 12

    #RECTANGLE, ELLIPSE, or TRIANGLE
    #  Use _obj_par1 to get ratio, or create 
    #  random ratio if _obj_par2 = True
    if (_obj_type == 1 or _obj_type == 2 or 
        _obj_type == 3): #rectangle, ellipse, or triangle
        if ((_obj_par1 != None and _obj_par2 == True) or 
            (_obj_par1 == None and _obj_par2 == True)):
            #randomize width to length ratio
            larger = get_dist(_min_value = _obj_size_min, 
                              _max_value = _obj_size_max, 
                              _dist = _size_dist)
            smaller = get_dist(_min_value = _obj_size_min, 
                               _max_value = _obj_size_max, 
                               _dist = _size_dist)             
        elif ((_obj_par1 != None and _obj_par2 == None) or 
              (_obj_par1 != None and _obj_par2 != None) or 
              (_obj_par1 != None and _obj_par2 == False)):
            #use provided smaller to larger ratio
            #  get random int between min and max for larger
            #  use ratio to get associated int smaller
            if _obj_size_min == _obj_size_max:
                larger = _obj_size_max
            else:
                larger = get_dist(_min_value = _obj_size_min, 
                                  _max_value = _obj_size_max, 
                                  _dist = _size_dist)
            smaller = int(larger*_obj_par1)
        else: 
            #define ratio as default
            #  get random int between min and max for length
            #  use default ratio to get assoceated smaller
            if _obj_size_min == _obj_size_max:
                larger = _obj_size_max
            else:
                larger = get_dist(_min_value = _obj_size_min, 
                                  _max_value = _obj_size_max, 
                                  _dist = _size_dist)
            smaller = int(larger*_default_axis_ratio)
        return smaller, larger

    #POLYGON
    #  Use _obj_par1 to get number of vertices, or create 
    #  random number of vertices if _obj_par2 = True
    if _obj_type == 4: #polygon
        #Check if _obj_par2 == True
        if _obj_par2 == True:
            if _obj_size_min == _obj_size_max:
                radius = _obj_size_min
            else:
                radius = get_dist(_min_value = _obj_size_min, 
                                  _max_value = _obj_size_max, 
                                  _dist = _size_dist)
            # use random number of polygon vertices.
            num_polygon_vertices = np.random.randint(min_number_sides, 
                                                     max_number_sides)
        #else, check if _obj_par1 == an intiger
        elif isinstance(_obj_par1, int):
            if _obj_size_min == _obj_size_max:
                radius = _obj_size_min
            else:
                radius = get_dist(_min_value = _obj_size_min, 
                                  _max_value = _obj_size_max, 
                                  _dist = _size_dist)
            # use user provided number of polygon vertices.
            num_polygon_vertices = _obj_par1
        #else, set sides to default
        else:
            if _obj_size_min == _obj_size_max:
                radius = _obj_size_min
            else:
                radius = get_dist(_min_value = _obj_size_min, 
                                  _max_value = _obj_size_max, 
                                  _dist = _size_dist)
            # use hexagon (all 6 sides same size)
            num_polygon_vertices = num_polygon_vertices
        return radius, num_polygon_vertices

    #RANDOM
    #  Select random shape from list of _obj_par1 shapes or default 
    #  (all shapes).
    #  1  = rectangle, 2 = ellipse, 3 = triangle, 4 = polygon 
    if _obj_type == 5: #random shapes
        #Turn _obj_par1 into a list if user failed to enter it as a list
        if not isinstance(_obj_par1, list): 
            if isinstance(_obj_par1, str):
                split_string = _obj_par1.split(",")
                _obj_par1 = [int(i) for i in split_string]
            if isinstance(_obj_par1, tuple):
                _obj_par1 = list(_obj_par1)                

        #Create list of possible shapes (in case user accidentally added 
        #    non-viable options to their list). If "shape" is in _obj_par1, 
        #    append "shapes" with shape number
        shape_options = []
        if (_obj_par1 != None and ((1 in _obj_par1) or
                                   (2 in _obj_par1) or
                                   (3 in _obj_par1) or
                                   (4 in _obj_par1))):
            if 1 in _obj_par1:
                shape_options.append(1) #rectangle
            if 2 in _obj_par1:
                shape_options.append(2) #ellipse
            if 3 in _obj_par1:
                shape_options.append(3) #triangle
            if 4 in _obj_par1:
                shape_options.append(4) #polygon
        else:
            #[1='rectangle',2='ellipse',3='triangle',4='polygon']
            shape_options = [1,2,3,4]
 
        #Get random shape from list
        #[1='rectangle',2='ellipse',3='triangle',4='polygon']
        random_shape = np.random.choice(shape_options)
        if random_shape == 1:
            return(1) #rectangle
        elif random_shape == 2:
            return(2) #ellipse
        elif random_shape == 3:
            return(3) #triangle
        elif random_shape == 4:
            return(4) #polygon
        else:
            return(2) #ellipse


# ********************** Rotate vertices of an object counterclockwise by a given angle
def rot_obj_vert(cx, cy, _rows, _columns, _ang_min = 0, _ang_max = 0, _ang_dist = 1): #RAC25032020
    '''
    Rotate vertices counterclockwise by a given angle (degrees) around center.
    Used for shapes: rectangle, square, polygon

    Rotation Parameters
        :param cx = x: coordinate of object center.
        :param cy = y: coordinate of object center.
        :param _rows: x coordinates of object vertices (array).
        :param _columns: y coordinates of object vertices (array).
        :param _ang_min = 0: minimum rotation angle of shape [degrees].
        :param _ang_max = 0: maximum rotation angle of shape [degrees].
        :param _ang_dist = 1: distribution of rotation angle. Choices are: 
            1=uniform, 2=normal(Gaussian), 3=Flory–Schulz
    '''

    try:
        _rows.size == _columns.size
    except:
        print("Can not rotate object. Number of row and column vertex coordinates are not equal")
        
    try:
        _ang_max >= _ang_min
    except:
        print("Maximum angle of rotation is larger than minimum angle of rotation")

    #Rotate vertex coordinates
    rotated_columns = [] #x values
    rotated_rows = [] #y values
    if _ang_max == _ang_min:
        for ii in range(_rows.size):
            _angle = _ang_max
            rotated_columns.append(int(cx + np.cos(np.deg2rad(_angle)) * 
                                       (_columns[ii] - cx) - np.sin(np.deg2rad(_angle)) * 
                                       (_rows[ii] - cy)))
            rotated_rows.append(int(cy + np.sin(np.deg2rad(_angle)) * 
                                    (_columns[ii] - cx) + np.cos(np.deg2rad(_angle)) * 
                                    (_rows[ii] - cy)))
    else:
        _angle = get_dist(_min_value = _ang_min, _max_value = _ang_max, _dist = _ang_dist)
        for ii in range(_rows.size): #Uniform Distribution Parameters
            rotated_columns.append(int(cx + np.cos(np.deg2rad(_angle)) * 
                                       (_columns[ii] - cx) - np.sin(np.deg2rad(_angle)) * 
                                       (_rows[ii] - cy)))
            rotated_rows.append(int(cy + np.sin(np.deg2rad(_angle)) * 
                                    (_columns[ii] - cx) + np.cos(np.deg2rad(_angle)) * 
                                    (_rows[ii] - cy)))

    return rotated_rows, rotated_columns


# ********************** Get value using a given distribution
def get_dist(_min_value, _max_value, _dist = 1): #RAC30032020
    '''Get value using a given distribution.

    Distribution Parameters
        :param _min_value = 0: minimum value.
        :param _max_value = 0: maximum value.
        :param _dist = 1: distribution of value selection. Choices are: 
            1=uniform, 2=normal(Gaussian), 3=Flory–Schulz
    '''

    # load package srwl_uti_dist
    try:
        sys.path.append(os.path.join(os.path.dirname(__file__), '..'))
        from uti_math import get_dist_uni, get_dist_norm, get_dist_schultz
    except:
        print('uti_math functions: get_uni_rand, get_norm_dist, get_schulz_dist failed to load.')

    # chose distribution method
    if _dist == 2: #Normal (Gaussian) Distribution Parameters
        value = get_dist_norm(_min_value, _max_value, _scale=((_max_value-_min_value)/3))
    elif _dist == 3: #Flory–Schulz Distribution Parameters
        value = get_dist_schultz(_min_value, _max_value)
    else: #Uniform Distribution Parameters
        value = get_dist_uni(_min_value, _max_value)  

    return value










#############################################################################
# Basic Input/Output Utility Functions
# v 0.02
#############################################################################

from copy import *

#**********************Auxiliary function to write auxiliary/debugging information to an ASCII file:
def write_text(_text, _file_path):
    f = open(_file_path, 'w')
    f.write(_text + '\n')
    f.close()

#**********************Auxiliary function to read-in data comumns from ASCII file (2D table):
def read_ascii_data_cols(_file_path, _str_sep, _i_col_start=0, _i_col_end=-1, _n_line_skip=0, _float=True, _n_line_skip_end=0): #OC20042020
#def read_ascii_data_cols(_file_path, _str_sep, _i_col_start=0, _i_col_end=-1, _n_line_skip=0, _float=True): #OC24112019
#def read_ascii_data_cols(_file_path, _str_sep, _i_col_start=0, _i_col_end=-1, _n_line_skip=0):
    """
    Auxiliary function to read-in data comumns from ASCII file (2D table)
    :param _file_path: full path (including file name) to the file
    :param _str_sep: column separation symbol(s) (string)
    :param _i_col_start: initial data column to read
    :param _i_col_end: final data column to read
    :param _n_line_skip: number of lines to skip in the beginning of the file
    :return: 2D list containing data columns read
    """
    f = open(_file_path, 'r')
    lines = f.readlines()

    resCols = []

    #nCol = _i_col_end - _i_col_start + 1
    #for iCol in range(nCol):
    #    resCols.append([])

    nRows = len(lines) - _n_line_skip - _n_line_skip_end #OC20042020
    #nRows = len(lines) - _n_line_skip

    for i in range(nRows):
        curLine = lines[_n_line_skip + i]
        curLineParts = curLine.split(_str_sep)
        curNumParts = len(curLineParts)
        #print(curLineParts)

        colCount = 0; colCountTrue = 0
        for iCol in range(curNumParts):
            curPart = curLineParts[iCol]
            #print(curPart)
            
            if(len(curPart) > 0):
                if(((_i_col_start <= colCount) or (_i_col_start < 0)) and ((colCount <= _i_col_end) or (_i_col_end < 0))):
                    if len(resCols) < (colCountTrue + 1): resCols.append([])
                    #resCols[colCountTrue].append(float(curPart))
                    if(_float): resCols[colCountTrue].append(float(curPart)) #OC24112019
                    else: resCols[colCountTrue].append(curPart)
                    colCountTrue += 1
                colCount += 1
    f.close()
    return resCols #attn: returns lists, not arrays!

#**********************Auxiliary function to write (save) data comumns to ASCII file (2D table):
def write_ascii_data_cols(_file_path, _cols, _str_sep, _str_head=None, _i_col_start=0, _i_col_end=-1):
    """
    Auxiliary function to write tabulated data (columns, i.e 2D table) to ASCII file
    :param _file_path: full path (including file name) to the file to be (over-)written
    :param _cols: array of data columns to be saved to file
    :param _str_sep: column separation symbol(s) (string)
    :param _str_head: header (string) to write before data columns
    :param _i_col_start: initial data column to write
    :param _i_col_end: final data column to write
    """
    f = open(_file_path, 'w')

    if(_str_head != None):
        lenStrHead = len(_str_head)
        if(lenStrHead > 0):
            strHead = _str_head
            if(_str_head[lenStrHead - 1] != '\n'):
                strHead = copy(_str_head) + '\n'
            f.write(strHead)
    if(_cols == None):
        f.close(); return
        
    nCols = len(_cols)
    if(nCols <= 0):
        f.close(); return

    nLines = len(_cols[0])
    for i in range(1, nCols):
        newLen = len(_cols[i])
        if(nLines < newLen): nLines = newLen

    strSep = '\t'
    if(_str_sep != None):
        if(len(_str_sep) > 0): strSep = _str_sep

    strTot = ''
    iColEndP1 = nCols
    if((_i_col_end >= 0) and (_i_col_end < nCols)): iColEndP1 = _i_col_end + 1
    iColEnd = iColEndP1 - 1
    nLinesM1 = nLines - 1
        
    for i in range(nLines):
        curLine = ''
        for j in range(_i_col_start, iColEndP1):
            curElem = ' '
            if(i < len(_cols[j])): curElem = repr(_cols[j][i])
            curLine += curElem
            if(j < iColEnd): curLine += strSep
        if(i < nLinesM1): curLine += '\n'
        strTot += curLine
        
    f.write(strTot)
    f.close()

#**********************Auxiliary function to write (save) data rows to ASCII file (2D table):
def write_ascii_data_rows(_file_path, _rows, _str_sep, _str_head=None, _i_col_start=0, _i_col_end=-1, _i_row_start=0, _i_row_end=-1): #OC16112019
    """
    Auxiliary function to write tabulated data (columns, i.e 2D table) to ASCII file
    :param _file_path: full path (including file name) to the file to be (over-)written
    :param _rows: array of data rows (/lines) to be saved to file
    :param _str_sep: column separation symbol(s) (string)
    :param _str_head: header (string) to write before data columns
    :param _i_col_start: initial data column to write
    :param _i_col_end: final data column to write
    :param _i_row_start: initial data row to write
    :param _i_row_end: final data row to write
    """
    f = open(_file_path, 'w')

    if(_str_head != None):
        lenStrHead = len(_str_head)
        if(lenStrHead > 0):
            strHead = _str_head
            if(_str_head[lenStrHead - 1] != '\n'):
                strHead = copy(_str_head) + '\n'
            f.write(strHead)
    if(_rows == None):
        f.close(); return
        
    nRows = len(_rows)
    nCols = len(_rows[0])

    if((nRows <= 0) or (nCols <= 0)):
        f.close(); return

    strSep = '\t'
    if(_str_sep != None):
        if(len(_str_sep) > 0): strSep = _str_sep

    strTot = ''
    iColEndP1 = nCols
    if((_i_col_end >= 0) and (_i_col_end < nCols)): iColEndP1 = _i_col_end + 1

    iRowEndP1 = nRows
    if((_i_row_end >= 0) and (_i_row_end < nRows)): iRowEndP1 = _i_row_end + 1

    iColEnd = iColEndP1 - 1
    iRowEnd = iRowEndP1 - 1
        
    for i in range(_i_row_start, iRowEndP1):
        curLine = ''
        curDataRow = _rows[i]
        for j in range(_i_col_start, iColEndP1):
            curElem = repr(curDataRow[j])
            curLine += curElem
            if(j < iColEnd): curLine += strSep
        if(i < iRowEnd): curLine += '\n'
        strTot += curLine
        
    f.write(strTot)
    f.close()

#********************** Read data from an image file:
def read_image(image_path, _do8bit=True):  #OC23102019
#def read_image(image_path):  # MR26102017
    """Read data from an image file.

    :param image_path: full path to the image.
    :return: dict with the processed data.
    """

    msg = '{0} library is not installed. Use "pip install {0}" to install it.'
    try:
        import numpy as np
    except:
        raise ImportError(msg.format('numpy'))
    try:
        from PIL import Image
    except:
        raise ImportError(msg.format('pillow'))

    #OC11112018 (as suggested by Rafael Celestre, to walk around image size limit)
    Image.MAX_IMAGE_PIXELS = None

    # Read the image:
    raw_image = Image.open(image_path)
    image_format = raw_image.format

    #OC23102019
    #raw_image = raw_image.convert('L')
    if(_do8bit): raw_image = raw_image.convert('L')

    # Convert it to NumPy array:
    data = np.array(raw_image)
    if image_format not in ('TIFF', 'PNG', 'BMP', 'GIF', 'JPEG'):
        raise ValueError('"{}" format is not supported at the moment.'.format(raw_image.format))

    # Get bits per point:
    mode_to_bpp = {'1': 1, 'L': 8, 'P': 8, 'I;16': 16, 'RGB': 24, 'RGBA': 32, 'CMYK': 32, 'YCbCr': 24, 'I': 32, 'F': 32}
    bpp = mode_to_bpp[raw_image.mode]
    limit_value = float(2 ** bpp - 1)

    return {
        'data': data,
        'raw_image': raw_image,
        'limit_value': limit_value,
    }



#############################################################################
# SRWLib for Python: Magnet Utilities v 0.02
#############################################################################

from srwlib import *
#from copy import *

#****************************************************************************
def srwl_mag_kick(_el_en=3., _ang=1., _x_or_y='x', _len=1., _led=0):
    """Setup dipole "kick" magnet
    :param _el_en: electron energy [GeV]
    :param _ang: kick angle [rad]
    :param _x_or_y: plane (horizontal or vertical)
    :param _len: effective magnet length [m]
    :param _led: "soft" edge length for field variation from 10% to 90% [m]; G/(1 + ((z-zc)/d)^2)^2 fringe field dependence is assumed [m]
    :param _zc: longitudinal position of the magnet center
    :param _add: add (=1) or reset (=0) the new magnet structure to the existing approximate magnetic field container
    """

    if(_el_en <= 0): raise Exception('Electron Beam structure is not defined (it is required for defining kick magnet parameters from angle)')
    if(_len <= 0): raise Exception('Inconsistent input magnet parameters: effective length can not be negative')

    if(_led > 0.):
        d = 0.8078259211948791*_led #d parameter in G/(1 + ((z-zc)/d)^2)^2
        L0 = _len - 1.5707963267948966*d #length of const. magnetic field part
        if(L0 < 0.): raise Exception('Inconsistent input magnet parameters: magnet edge length if too large for given effective length')

    B = -3.33564095*_ang*_el_en/_len #sign to be checked!
    n_or_s = 'n'
    if(_x_or_y == 'y'): n_or_s = 's' #to check
    
    return SRWLMagFldM(_G=B, _m=1, _n_or_s=n_or_s, _Leff=_len, _Ledge=_led)

#****************************************************************************
def srwl_mag_extrap_grad_off_mid_plane(_mag_mid, _ry, _ny, _grad_mult=1):
    """
    Extrapolates magnetic field off horizontal mid-plane based on field gradient in the mid-plane (e.g. for dipole with gradient or for a guad).
    :param _mag_mid: input tabulated 3D magnetic field on 2D mesh vs x and z in the horizontal mid-plane in a container (object of SRWLMagFldC type)
    :param _ry: vertical position range of the final 3D mesh [m]
    :param _ny: number of points vs vertical position in the final 3D mesh
    :param _grad_mult: a number the gradient has to be multiplied by
    :returns: resulting exrapolated 3D magnetic field structure in a container (object of SRWLMagFldC type)
    """

    if((_mag_mid is None) or (_mag_mid.arMagFld is None) or (len(_mag_mid.arMagFld) != 1) or (_ry < 0.) or (_ny <= 1)):
        raise Exception("Incorrect input parameters for magnetic field to be extrapolated")

    fld3d_mid = _mag_mid.arMagFld[0]
    #arBx_mid = fld3d_mid.arBx
    arBy_mid = fld3d_mid.arBy
    arBz_mid = fld3d_mid.arBz

    xc = _mag_mid.arXc[0]
    xStart = xc - 0.5*fld3d_mid.rx
    nx = fld3d_mid.nx
    xStep = fld3d_mid.rx/(nx - 1)

    yc = 0
    yStart = -0.5*_ry
    ny = int(_ny)
    yStep = _ry/(ny - 1)

    zc = _mag_mid.arZc[0]
    zStart = zc - 0.5*fld3d_mid.rz
    nz = fld3d_mid.nz
    zStep = fld3d_mid.rz/(nz - 1)

    nTotRes = int(nx*ny*nz)
    arBxRes = array('d', [0]*nTotRes)
    arByRes = array('d', [0]*nTotRes)
    arBzRes = None if(arBz_mid is None) else array('d', [0]*nTotRes)
    
    nx_mi_1 = nx - 1
    #print(nx, xStart, xStep, ny, yStart, yStep, nz, zStart, zStep)
    
    for iz in range(nz):
        iz_nx = iz*nx
        iz_nx_ny = iz*nx*ny
        for ix in range(nx):
            b1y = 0; b2y = 0; b0y = 0; dx = xStep
            if(ix == 0):
                b1y = arBy_mid[iz_nx]
                b2y = arBy_mid[iz_nx + 1]
                b0y = b1y
            elif(ix == nx_mi_1):
                b1y = arBy_mid[iz_nx + nx_mi_1 - 1]
                b2y = arBy_mid[iz_nx + nx_mi_1]
                b0y = b2y
            else:
                b1y = arBy_mid[iz_nx + ix - 1]
                b0y = arBy_mid[iz_nx + ix]
                b2y = arBy_mid[iz_nx + ix + 1]
                dx = 2*xStep
                
            curGrad = _grad_mult*(b2y - b1y)/dx
            y = yStart
            for iy in range(ny):
                ofst = iz_nx_ny + iy*nx + ix
                arBxRes[ofst] = y*curGrad
                arByRes[ofst] = b0y
                y += yStep

    fld3dRes = SRWLMagFld3D(arBxRes, arByRes, arBzRes, nx, ny, nz, fld3d_mid.rx, _ry, fld3d_mid.rz)
    #print(fld3dRes.nx, fld3dRes.rx, fld3dRes.ny, fld3dRes.ry, fld3dRes.nz, fld3dRes.rz)

    return SRWLMagFldC(fld3dRes, xc, yc, zc)

#****************************************************************************
def srwl_mag_track_e_beam_mom(_e_beam, _mag, _z_or_ct, _ct_start=0, _ct_end=0, _npart=1000, _np_in_trj=10000):
    """Estimate Electron Beam Statistical Moments at a given long. pos. (for tests)
    :param _e_beam: input electron beam structure (object of SRWLPartBeam type)
    :param _mag: input magnetic field container (object of SRWLMagFldC type)
    :param _z_or_ct: longitudinal position [m] where the stat. moments have to be calculated, or c*t (if _ct_start != _ct_end)
    :param _ct_start: start time ct moment [m] for trajectory calculation
    :param _ct_end: end timect  moment [m] for trajectory calculation
    :param _npart: number of macro-particles
    :param _np_in_trj: number of points in trajectory
    """

    partTraj = SRWLPrtTrj()
    partTraj.allocate(_np_in_trj, True)
    partTraj.ctStart = _ct_start #Start Time for the calculation
    partTraj.ctEnd = _ct_end #magFldCnt.arMagFld[0].rz

    x0 = _e_beam.partStatMom1.x
    xp0 = _e_beam.partStatMom1.xp
    y0 = _e_beam.partStatMom1.y
    yp0 = _e_beam.partStatMom1.yp
    #en0GeV = _e_beam.partStatMom1.get_E()
    gamma0 = _e_beam.partStatMom1.gamma

    sigXe2 = _e_beam.arStatMom2[0] #[0]: <(x-x0)^2>
    mXXp = _e_beam.arStatMom2[1] #[1]: <(x-x0)*(xp-xp0)>
    sigXpe2 = _e_beam.arStatMom2[2] #[2]: <(xp-xp0)^2>
    sigYe2 = _e_beam.arStatMom2[3] #[3]: <(y-y0)^2>
    mYYp = _e_beam.arStatMom2[4] #[4]: <(y-y0)*(yp-yp0)>
    sigYpe2 = _e_beam.arStatMom2[5] #[5]: <(yp-yp0)^2>
    relEnSpr = sqrt(_e_beam.arStatMom2[10]) #<(E-E0)^2>/E0^2
    #absEnSpr = en0GeV*relEnSpr

    multX = 0.5/(sigXe2*sigXpe2 - mXXp*mXXp)
    BX = sigXe2*multX
    GX = sigXpe2*multX
    AX = mXXp*multX
    sigPX = 1/sqrt(2*GX)
    sigQX = sqrt(sigXpe2)

    multY = 0.5/(sigYe2*sigYpe2 - mYYp*mYYp)
    BY = sigYe2*multY
    GY = sigYpe2*multY
    AY = mYYp*multY
    sigPY = 1/sqrt(2*GY)
    sigQY = sqrt(sigYpe2)

    randAr = array('d', [0]*5) #for random Gaussian numbers
    random.seed(12345)

    ind0 = 0
    r0 = 0
    ctStep = 0
    treatCT = False
    np_in_trj_mi_1 = _np_in_trj - 1
    if(_ct_start != _ct_end): #treat _z_or_ct as ct
        ctStep = (_ct_end - _ct_start)/np_in_trj_mi_1
        ind0 = int((_z_or_ct - _ct_start)/ctStep)
        if((ind0 >= 0) and (ind0 < np_in_trj_mi_1)): 
            r0 = (_z_or_ct - (_ct_start + ind0*ctStep))/ctStep
            treatCT = True
        if(treatCT is not True):
            raise Exception("Trajectory point corresponding to given time moment was not found")

    xSum = 0; ySum = 0
    xpSum = 0; ypSum = 0
    xe2Sum = 0; ye2Sum = 0
    xpe2Sum = 0; ype2Sum = 0
    xxpSum = 0; yypSum = 0
    for i in range(_npart):

        for ir in range(5):
            randAr[ir] = random.gauss(0, 1)

        auxPXp = sigQX*randAr[0]
        auxPX = sigPX*randAr[1] + AX*auxPXp/GX
        partTraj.partInitCond.x = x0 + auxPX
        partTraj.partInitCond.xp = xp0 + auxPXp
        auxPYp = sigQY*randAr[2]
        auxPY = sigPY*randAr[3] + AY*auxPYp/GY
        partTraj.partInitCond.y = y0 + auxPY
        partTraj.partInitCond.yp = yp0 + auxPYp
        partTraj.partInitCond.gamma = gamma0*(1 + relEnSpr*randAr[4])

        partTraj = srwl.CalcPartTraj(partTraj, _mag, [1])

        if treatCT is False:
            #Search for longitudinal position:
            iz = ind0
            dZprev = 1.e+23
            newIndFound = False
            while(iz >= 0):
                if(iz < np_in_trj_mi_1):
                    if((partTraj.arZ[iz] <= _z_or_ct) and (_z_or_ct < partTraj.arZ[iz + 1])):
                        ind0 = iz
                        newIndFound = True
                        break
                dZ = abs(_z_or_ct - partTraj.arZ[iz])
                if(dZ > dZprev): break

                dZprev = dZ
                iz -= 1

            if newIndFound is False:
                iz = ind0
                dZprev = 1.e+23
                while(iz < np_in_trj_mi_1):
                    if((partTraj.arZ[iz] <= _z_or_ct) and (_z_or_ct < partTraj.arZ[iz + 1])):
                        ind0 = iz
                        newIndFound = True
                        break
                    dZ = abs(_z_or_ct - partTraj.arZ[iz])
                    if(dZ > dZprev): break

                    dZprev = dZ
                    iz += 1

            if(newIndFound is False):
                raise Exception("Trajectory point corresponding to given longitudinal position was not found")

            r0 = (_z_or_ct - partTraj.arZ[ind0])/(partTraj.arZ[ind0 + 1] - partTraj.arZ[ind0])

        x1 = partTraj.arX[ind0]; x2 = partTraj.arX[ind0 + 1]
        x = x1 + r0*(x2 - x1)
        xp1 = partTraj.arXp[ind0]; xp2 = partTraj.arXp[ind0 + 1]
        xp = xp1 + r0*(xp2 - xp1)
        y1 = partTraj.arY[ind0]; y2 = partTraj.arY[ind0 + 1]
        y = y1 + r0*(y2 - y1)
        yp1 = partTraj.arYp[ind0]; yp2 = partTraj.arYp[ind0 + 1]
        yp = yp1 + r0*(yp2 - yp1)

        xSum += x; ySum += y
        xpSum += xp; ypSum += yp
        xe2Sum += x*x; ye2Sum += y*y
        xpe2Sum += xp*xp; ype2Sum += yp*yp
        xxpSum += x*xp; yypSum += y*yp
    
    invN = 1./_npart
    xAvg = xSum*invN; yAvg = ySum*invN
    xpAvg = xpSum*invN; ypAvg = ypSum*invN
    xe2Avg = xe2Sum*invN; ye2Avg = ye2Sum*invN
    xpe2Avg = xpe2Sum*invN; ype2Avg = ype2Sum*invN
    xxpAvg = xxpSum*invN; yypAvg = yypSum*invN

    sigXe2 = xe2Avg - xAvg*xAvg; sigYe2 = ye2Avg - yAvg*yAvg
    mXXp = xxpAvg - xAvg*xpAvg; mYYp = yypAvg - yAvg*ypAvg
    sigXpe2 = xpe2Avg - xpAvg*xpAvg; sigYpe2 = ype2Avg - ypAvg*ypAvg

    return [[xAvg, xpAvg, sigXe2, mXXp, sigXpe2], [yAvg, ypAvg, sigYe2, mYYp, sigYpe2]]

#****************************************************************************
##def srwl_mag_extrap_grad_curv(_mag_curv_mid, _cen_trj_data, _xi, _xf, _nx, _yi, _yf, _ny, _zi, _zf, _nz, _grad_mult=1):
##    """
##    Extrapolates magnetic field on rectangulat 3D mesh in and off horizontal mid-plane based on field gradient in the mid-plane (e.g. for dipole with gradient or for a guad).
##    :param _mag_curv_mid: input tabulated 3D magnetic field on 2D mesh vs x and z in Natural frame in the horizontal mid-plane in a container (object of SRWLMagFldC type)
##    :param _cen_trj_data: central trajectory data defining the Natural frame of the input magnetic field (_cen_trj_data[0] is ct, _cen_trj_data[1] is x. _cen_trj_data[2] is z)
##    :param _xi: horizontal initial position of the final 3D mesh [m]
##    :param _xf: horizontal final position of the final 3D mesh [m]
##    :param _nx: number of points vs horizontal position in the final 3D mesh
##    :param _yi: vertical initial position of the final 3D mesh [m]
##    :param _yf: vertical initial position of the final 3D mesh [m]
##    :param _ny: number of points vs vertical position in the final 3D mesh
##    :param _zi: longitudinal initial position of the final 3D mesh [m]
##    :param _zf: longitudinal final position of the final 3D mesh [m]
##    :param _nz: number of points vs longitudinal position in the final 3D mesh
##    :param _grad_mult: a number the gradient has to be multiplied by
##    :returns: exrapolated 3D magnetic field structure in a container (object of SRWLMagFldC type)
##    """
##
##    if((_mag_curv_mid is None) or (_mag_curv_mid.arMagFld is None) or (len(_mag_curv_mid.arMagFld) < 1) or (_cen_trj_data is None) or (len(_cen_trj_data) < 3)
##       or (_nx < 1) or (_ny < 1) or (_nz < 1)):
##        raise Exception("Incorrect input parameters for magnetic field to be extrapolated")
##    #print('In srwl_fld_extrap_grad_curv')
##
##    fld3d_curv_mid = _mag_curv_mid.arMagFld[0]
##    #arBx_curv_mid = fld3d_curv_mid.arBx
##    arBy_curv_mid = fld3d_curv_mid.arBy
##    arBz_curv_mid = fld3d_curv_mid.arBz



#############################################################################
# uti_math module: misc. mathematical utilities / functions
# v 0.04
# Authors: O.C., Maksim Rakitin, Rebecca Ann Coles
#############################################################################

from __future__ import print_function #Python 2.7 compatibility
from array import *
from math import *
from copy import *
import random
#import sys
#import os

#****************************************************************************
def interp_1d(_x, _x_min, _x_step, _nx, _ar_f, _ord=3, _ix_per=1, _ix_ofst=0):
    """
    Interpolate 1D function value tabulated on equidistant mesh, using polynomial interpolation
    :param _x: argument at which function value should be calculated
    :param _x_min: minimal argument value of the tabulated function
    :param _x_step: step of mesh at which function is tabulated
    :param _nx: number of points in mesh at which function is tabulated
    :param _ar_f: tabulated function list or array
    :param _ord: order of polynomial interpolation (1- linear, 2- quadratic, 3- cubic)
    :param _ix_per: argument index period of function data alignment (e.g. to interpolate one component of complex data, or in one dimension of multi-dimensional data)
    :param _ix_ofst: argument index offset of function data alignment
    :return: function value found by polynomial interpolation
    """
    if(_ord == 1):
        i0 = int(trunc((_x - _x_min)/_x_step + 1.e-09))
        if(i0 < 0):
            i0 = 0
        elif(i0 >= _nx - 1):
            i0 = _nx - 2
        i1 = i0 + 1
        f0 = _ar_f[i0*_ix_per + _ix_ofst]
        f1 = _ar_f[i1*_ix_per + _ix_ofst]
        t = (_x - (_x_min + _x_step*i0))/_x_step
        return f0 + (f1 - f0)*t
    elif(_ord == 2):
        i0 = int(round((_x - _x_min)/_x_step))
        if(i0 < 1):
            i0 = 1
        elif(i0 >= _nx - 1):
            i0 = _nx - 2
        im1 = i0 - 1
        i1 = i0 + 1
        t = (_x - (_x_min + _x_step*i0))/_x_step
        a0 = _ar_f[i0*_ix_per + _ix_ofst]
        fm1 = _ar_f[im1*_ix_per + _ix_ofst]
        f1 = _ar_f[i1*_ix_per + _ix_ofst]
        a1 = 0.5*(f1 - fm1)
        a2 = 0.5*(fm1 + f1 - 2*a0)
        return a0 + t*(a1 + t*a2)
    elif(_ord == 3):
        i0 = int(trunc((_x - _x_min)/_x_step + 1.e-09))
        if(i0 < 1):
            i0 = 1
        elif(i0 >= _nx - 2):
            i0 = _nx - 3
        im1 = i0 - 1
        i1 = i0 + 1
        i2 = i0 + 2
        t = (_x - (_x_min + _x_step*i0))/_x_step
        a0 = _ar_f[i0*_ix_per + _ix_ofst]
        fm1 = _ar_f[im1*_ix_per + _ix_ofst]
        f1 = _ar_f[i1*_ix_per + _ix_ofst]
        f2 = _ar_f[i2*_ix_per + _ix_ofst]
        a1 = -0.5*a0 + f1 - f2/6. - fm1/3.
        a2 = -a0 + 0.5*(f1 + fm1)
        a3 = 0.5*(a0 - f1) + (f2 - fm1)/6.
        return a0 + t*(a1 + t*(a2 + t*a3))
    return 0

#****************************************************************************
#def interp_1d_lin_var(_x, _ar_x, _ar_f):
def interp_1d_var(_x, _ar_x, _ar_f, _ord=3):
    """
    Interpolate linearly 1D function value tabulated on non-equidistant (irregular) mesh
    :param _x: argument at which function value should be calculated
    :param _ar_x: array or list of increasing argument values, at which the function is tabulated
    :param _ar_f: array or list of tabulated values corresponding to arguments in _ar_x
    :param _ord: order of polynomial interpolation (1- linear, 2- quadratic, 3- cubic)    
    :return: function value found by polynomial interpolation
    """

    sErrBadArrays = 'Incorrect/incompatible lengths of argument and function value arrays.'

    nx = len(_ar_x)
    if(nx <= 0): raise Exception(sErrBadArrays)
    nf = len(_ar_f)
    if(nf <= 0): raise Exception(sErrBadArrays)
    
    if(nx > nf): nx = nf
    if(nx < 2): raise Exception(sErrBadArrays)
    
    nx_mi_1 = nx - 1

    if(_x <= _ar_x[0]): return _ar_f[0]
    elif(_x >= _ar_x[nx_mi_1]): return _ar_f[nx_mi_1]

    if(_ord > nx_mi_1): _ord = nx_mi_1

    i0 = 0
    for i in range(1, nx):
        if(_x < _ar_x[i]):
            i0 = i - 1
            break

    if(_ord == 1):
        #return interp_1d(_x, _ar_x[i0], _ar_x[i0+1] - _ar_x[i0], 2, [_ar_f[i0], _ar_f[i0+1]], 1)
        x0 = _ar_x[i0]
        step = _ar_x[i0+1] - x0
        t = (_x - x0)/step
        f0 = _ar_f[i0]
        return f0 + (_ar_f[i0+1] - f0)*t

    elif(_ord == 2):
        im1 = i0 - 1
        ip1 = i0 + 1
        #nm1 = nx - 1
        if(ip1 < 0):
            im1 = 0
            i0 = 1
            ip1 = 2
        elif(ip1 > nx_mi_1):
            im1 = nx - 3
            i0 = nx - 2
            ip1 = nx_mi_1

        xm1 = _ar_x[im1]
        x0 = _ar_x[i0]
        xp1 = _ar_x[ip1]
        dxm1 = abs(_x - xm1)
        dx0 = abs(_x - x0)
        dxp1 = abs(_x - xp1)
        if(dxm1 < dx0):
            if(im1 > 0):
                im1 -= 1
                i0 -= 1
                ip1 -= 1
        elif(dxp1 < dx0):
            if(ip1 < nx_mi_1):
                im1 += 1
                i0 += 1
                ip1 += 1

        x0 = _ar_x[i0]
        dxm1 = _ar_x[im1] - x0
        dxp1 = _ar_x[ip1] - x0
        fm1 = _ar_f[im1]
        f0 = _ar_f[i0]
        fp1 = _ar_f[ip1]

        invD = 1./(dxm1* dxp1*(dxm1 - dxp1))
        a = ((dxm1 - dxp1)*f0 + dxp1*fm1 - dxm1*fp1)*invD
        b = (dxp1*dxp1*(f0 - fm1) + dxm1*dxm1*(fp1 - f0))*invD
        dx = _x - x0
        return (a*dx + b)*dx + f0

    elif(_ord == 3):
        im1 = i0 - 1
        ip1 = i0 + 1
        ip2 = i0 + 2
        if(im1 < 0):
            im1 = 0
            i0 = 1
            ip1 = 2
            ip2 = 3
        elif(ip2 > nx_mi_1):
            im1 = nx - 4
            i0 = nx - 3
            ip1 = nx - 2
            ip2 = nx_mi_1
            
        x0 = _ar_x[i0]
        dxm1 = _ar_x[im1] - x0
        dxp1 = _ar_x[ip1] - x0
        dxp2 = _ar_x[ip2] - x0
        fm1 = _ar_f[im1]
        f0 = _ar_f[i0]
        fp1 = _ar_f[ip1]
        fp2 = _ar_f[ip2]
        #print(_x - x0, dxm1, dxp1, dxp2)
        #print(fm1, f0, fp1, fp2)

        invD = 1./(dxm1*dxp1*dxp2*(dxm1 - dxp1)*(dxm1 - dxp2)*(dxp1 - dxp2))
        invD1 = 1./(dxm1*dxp1*dxp2)
        dxm1e2 = dxm1*dxm1
        dxm1e3 = dxm1e2*dxm1
        dxp1e2 = dxp1*dxp1
        dxp1e3 = dxp1e2*dxp1
        dxp2e2 = dxp2*dxp2
        dxp2e3 = dxp2e2*dxp2
        a1 = (-dxp1e2*(dxp1 - dxp2)*dxp2e2*(f0 - fm1) + dxm1e2*(dxp2e3*(fp1 - f0) + dxp1e3*(f0 - fp2)) + dxm1e3*(dxp2e2*(f0 - fp1) + dxp1e2*(fp2 - f0)))*invD
        a2 = ((dxm1 + dxp1 + dxp2)*f0)*invD1 + (dxm1*dxp2*(dxm1e2 - dxp2e2)*fp1 + dxp1e3*(dxm1*fp2 - dxp2*fm1) + dxp1*(dxp2e3*fm1 - dxm1e3*fp2))*invD
        a3 = -f0*invD1 + (dxm1*dxp2*(dxp2 - dxm1)*fp1 + dxp1e2*(dxp2*fm1 - dxm1*fp2) + dxp1*(dxm1e2*fp2 - dxp2e2*fm1))*invD
        dx = _x - x0
        return ((a3*dx + a2)*dx + a1)*dx + f0

#****************************************************************************
def interp_2d(_x, _y, _x_min, _x_step, _nx, _y_min, _y_step, _ny, _ar_f, _ord=3, _ix_per=1, _ix_ofst=0):
    """
    Interpolate 2D function value tabulated on equidistant rectangular mesh and represented by C-aligned flat array, using polynomial interpolation
    :param _x: first argument at which function value should be calculated
    :param _y: second argument at which function value should be calculated
    :param _x_min: minimal value of the first argument of the tabulated function
    :param _x_step: step of the first argument at which function is tabulated
    :param _nx: number of points vs first argument at which function is tabulated
    :param _y_min: minimal value of the second argument of the tabulated function
    :param _y_step: step of the second argument at which function is tabulated
    :param _ny: number of points vs second argument at which function is tabulated
    :param _ar_f: function tabulated on 2D mesh, aligned as "flat" C-type list or array (first argument is changing most frequently)
    :param _ord: "order" of polynomial interpolation (1- bi-linear (on 4 points), 2- "bi-quadratic" (on 6 points), 3- "bi-cubic" (on 12 points))
    :param _ix_per: period of first argument index of the function data alignment (e.g. to interpolate one component of complex data, or in one dimension of multi-dimensional data)
    :param _ix_ofst: offset of the first argument index in function data alignment
    :return: function value found by 2D polynomial interpolation
    """
    if(_ord == 1): #bi-linear interpolation based on 4 points
        ix0 = int(trunc((_x - _x_min)/_x_step + 1.e-09))
        if(ix0 < 0):
            ix0 = 0
        elif(ix0 >= _nx - 1):
            ix0 = _nx - 2
        ix1 = ix0 + 1
        tx = (_x - (_x_min + _x_step*ix0))/_x_step
        
        iy0 = int(trunc((_y - _y_min)/_y_step + 1.e-09))
        if(iy0 < 0):
            iy0 = 0
        elif(iy0 >= _ny - 1):
            iy0 = _ny - 2
        iy1 = iy0 + 1
        ty = (_y - (_y_min + _y_step*iy0))/_y_step

        nx_ix_per = _nx*_ix_per
        iy0_nx_ix_per = iy0*nx_ix_per
        iy1_nx_ix_per = iy1*nx_ix_per
        ix0_ix_per_p_ix_ofst = ix0*_ix_per + _ix_ofst
        ix1_ix_per_p_ix_ofst = ix1*_ix_per + _ix_ofst
        a00 = _ar_f[iy0_nx_ix_per + ix0_ix_per_p_ix_ofst]
        f10 = _ar_f[iy0_nx_ix_per + ix1_ix_per_p_ix_ofst]
        f01 = _ar_f[iy1_nx_ix_per + ix0_ix_per_p_ix_ofst]
        f11 = _ar_f[iy1_nx_ix_per + ix1_ix_per_p_ix_ofst]
        a10 = f10 - a00
        a01 = f01 - a00
        a11 = a00 - f01 - f10 + f11
        return a00 + tx*(a10 + ty*a11) + ty*a01

    elif(_ord == 2): #bi-quadratic interpolation based on 6 points
        ix0 = int(round((_x - _x_min)/_x_step))
        if(ix0 < 1):
            ix0 = 1
        elif(ix0 >= _nx - 1):
            ix0 = _nx - 2
        ixm1 = ix0 - 1
        ix1 = ix0 + 1
        tx = (_x - (_x_min + _x_step*ix0))/_x_step

        iy0 = int(round((_y - _y_min)/_y_step))
        if(iy0 < 1):
            iy0 = 1
        elif(iy0 >= _ny - 1):
            iy0 = _ny - 2
        iym1 = iy0 - 1
        iy1 = iy0 + 1
        ty = (_y - (_y_min + _y_step*iy0))/_y_step

        nx_ix_per = _nx*_ix_per
        iym1_nx_ix_per = iym1*nx_ix_per
        iy0_nx_ix_per = iy0*nx_ix_per
        iy1_nx_ix_per = iy1*nx_ix_per
        ixm1_ix_per_p_ix_ofst = ixm1*_ix_per + _ix_ofst
        ix0_ix_per_p_ix_ofst = ix0*_ix_per + _ix_ofst
        ix1_ix_per_p_ix_ofst = ix1*_ix_per + _ix_ofst
        fm10 = _ar_f[iy0_nx_ix_per + ixm1_ix_per_p_ix_ofst]
        a00 = _ar_f[iy0_nx_ix_per + ix0_ix_per_p_ix_ofst]
        f10 = _ar_f[iy0_nx_ix_per + ix1_ix_per_p_ix_ofst]
        f0m1 = _ar_f[iym1_nx_ix_per + ix0_ix_per_p_ix_ofst]
        f01 = _ar_f[iy1_nx_ix_per + ix0_ix_per_p_ix_ofst]
        f11 = _ar_f[iy1_nx_ix_per + ix1_ix_per_p_ix_ofst]
        a10 = 0.5*(f10 - fm10)
        a01 = 0.5*(f01 - f0m1)
        a11 = a00 - f01 - f10 + f11
        a20 = 0.5*(f10 + fm10) - a00
        a02 = 0.5*(f01 + f0m1) - a00
        return a00 + tx*(a10 + tx*a20 + ty*a11) + ty*(a01 + ty*a02)
    
    elif(_ord == 3): #bi-cubic interpolation based on 12 points
        ix0 = int(trunc((_x - _x_min)/_x_step + 1.e-09))
        if(ix0 < 1):
            ix0 = 1
        elif(ix0 >= _nx - 2):
            ix0 = _nx - 3
        ixm1 = ix0 - 1
        ix1 = ix0 + 1
        ix2 = ix0 + 2
        tx = (_x - (_x_min + _x_step*ix0))/_x_step

        iy0 = int(trunc((_y - _y_min)/_y_step + 1.e-09))
        if(iy0 < 1):
            iy0 = 1
        elif(iy0 >= _ny - 2):
            iy0 = _ny - 3
        iym1 = iy0 - 1
        iy1 = iy0 + 1
        iy2 = iy0 + 2
        ty = (_y - (_y_min + _y_step*iy0))/_y_step

        nx_ix_per = _nx*_ix_per
        iym1_nx_ix_per = iym1*nx_ix_per
        iy0_nx_ix_per = iy0*nx_ix_per
        iy1_nx_ix_per = iy1*nx_ix_per
        iy2_nx_ix_per = iy2*nx_ix_per
        ixm1_ix_per_p_ix_ofst = ixm1*_ix_per + _ix_ofst
        ix0_ix_per_p_ix_ofst = ix0*_ix_per + _ix_ofst
        ix1_ix_per_p_ix_ofst = ix1*_ix_per + _ix_ofst
        ix2_ix_per_p_ix_ofst = ix2*_ix_per + _ix_ofst
        f0m1 = _ar_f[iym1_nx_ix_per + ix0_ix_per_p_ix_ofst]
        f1m1 = _ar_f[iym1_nx_ix_per + ix1_ix_per_p_ix_ofst]
        fm10 = _ar_f[iy0_nx_ix_per + ixm1_ix_per_p_ix_ofst]
        a00 = _ar_f[iy0_nx_ix_per + ix0_ix_per_p_ix_ofst]
        f10 = _ar_f[iy0_nx_ix_per + ix1_ix_per_p_ix_ofst]
        f20 = _ar_f[iy0_nx_ix_per + ix2_ix_per_p_ix_ofst]
        fm11 = _ar_f[iy1_nx_ix_per + ixm1_ix_per_p_ix_ofst]
        f01 = _ar_f[iy1_nx_ix_per + ix0_ix_per_p_ix_ofst]
        f11 = _ar_f[iy1_nx_ix_per + ix1_ix_per_p_ix_ofst]
        f21 = _ar_f[iy1_nx_ix_per + ix2_ix_per_p_ix_ofst]
        f02 = _ar_f[iy2_nx_ix_per + ix0_ix_per_p_ix_ofst]
        f12 = _ar_f[iy2_nx_ix_per + ix1_ix_per_p_ix_ofst]
        a10 = -0.5*a00 + f10 - f20/6 - fm10/3
        a01 = -0.5*a00 + f01 - f02/6 - f0m1/3
        a11 = -0.5*(f01 + f10) + (f02 - f12 + f20 - f21)/6 + (f0m1 - f1m1 + fm10 - fm11)/3 + f11
        a20 = -a00 + 0.5*(f10 + fm10)
        a02 = -a00 + 0.5*(f01 + f0m1)
        a21 = a00 - f01 + 0.5*(f11 - f10 - fm10 + fm11)
        a12 = a00 - f10 + 0.5*(f11 - f01 - f0m1 + f1m1)
        a30 = 0.5*(a00 - f10) + (f20 - fm10)/6
        a03 = 0.5*(a00 - f01) + (f02 - f0m1)/6
        a31 = 0.5*(f01 + f10 - f11 - a00) + (f21 + fm10 - f20 - fm11)/6
        a13 = 0.5*(f10 - f11 - a00 + f01) + (f0m1 + f12 - f02 - f1m1)/6
        return a00 + tx*(a10 + tx*(a20 + tx*(a30 + ty*a31) + ty*a21) + ty*a11) + ty*(a01 + ty*(a02 + ty*(a03 + tx*a13) + tx*a12))
    return 0

#****************************************************************************
def num_round(_x, _ndig=8):
##    if(_x == 0.): return _x
##    sgn = 1.
##    if(_x < 0.):
##        _x = -_x
##        sgn = -1
##    order = round(log10(_x))
##    fact = 10**order
##    roundNum = round(_x/fact, _ndig)
##    res = roundNum*fact*sgn
    res = round(_x, _ndig)
    return res

#****************************************************************************
def find_ar_max(_ar, _ib=0, _ie=-1, _min=False):
    """
    Finds array (or list) maximum (or minimum), index and value 
    :param _ar: array (or list) to find max. or min.
    :param _ib: array index to start search
    :param _ie: array index to finish search
    :param _min: switch specifying that minimum (rather than maximum) has to be searched
    """

    strErIncorArray = 'Incorrect input array.'
    
    if(_ar is None): raise Exception(strErIncorArray)
    nTot = len(_ar)
    if(nTot <= 0): raise Exception(strErIncorArray)

    iBeg = _ib
    if(_ib < 0): iBeg = 0

    iEnd = _ie
    if((iEnd == -1) or (_ie >= nTot)): iEnd = nTot - 1

    if(iEnd < iBeg): raise Exception('Incorrect definition of start and end indexes.')

    curExtr = _ar[0]
    curInd = 0
    for i in range(iBeg, iEnd + 1, 1):
        curVal = _ar[i]
        if(_min == True): curVal *= -1
        if(curExtr < curVal): 
            curExtr = curVal
            curInd = i

    return curExtr, curInd

#****************************************************************************
def integ_array(_ar, _h, _dupl=False):
    """
    Integrates array (or list), eventually making a copy of it before the integration
    :param _ar: array to integrate
    :param _h: step size
    :param _dupl: duplicate the magnetic field object or not
    """
    ar = None
    if(_dupl): ar = deepcopy(_ar)
    else: ar = _ar

    hd2 = 0.5*_h
    auxInt = 0
    lenAr = len(_ar)
    for i in range(lenAr - 1):
        ar_i = _ar[i] #in case if array has to be integrated in place
        ar[i] = auxInt
        auxInt += hd2*(ar_i + _ar[i + 1])
    ar[lenAr - 1] = auxInt
    return ar

#****************************************************************************
def integ_ar_2d(_ar, _ar_align, _x_grid, _y_grid, _x_lim=None, _y_lim=None):
    """
    Integrates 2d array (or list) within given limits
    :param _ar: input array to integrate
    :param _ar_align: input array alignment (1- _ar is C-type alignment, 2- _ar is 2d array)
    :param _x_grid: list/array specifying grid vs one dimensions (_x_grid[0] is start, _x_grid[1] is end, _x_grid[2] is number of points)
    :param _y_grid: list/array specifying grid vs another dimensions (_y_grid[0] is start, _y_grid[1] is end, _y_grid[2] is number of points)
    :param _x_lim: list/array specifying inegration limits vs one dimensions (_x_lim[0] is start, _x_lim[1] is end)
    :param _y_lim: list/array specifying inegration limits vs another dimensions (_y_lim[0] is start, _y_lim[1] is end)
    """

    if(_x_lim is not None):
        #print(_x_lim[0], _x_lim[1])
        if(_x_lim[0] >= _x_lim[1]): return 0.

    if(_y_lim is not None):
        #print(_y_lim[0], _y_lim[1])
        if(_y_lim[0] >= _y_lim[1]): return 0.

    #print(_x_grid); print(_x_lim)
    #print(_y_grid); print(_y_lim)
    
    xStart = _x_grid[0]; xEnd = _x_grid[1]; nx = _x_grid[2]
    xStep = (xEnd - xStart)/(nx - 1)
    yStart = _y_grid[0]; yEnd = _y_grid[1]; ny = _y_grid[2]
    yStep = (yEnd - yStart)/(ny - 1)

    if((xStep == 0) or (yStep == 0)): return 0.

    x_min = xStart; x_max = xEnd; nxInteg = 0
    if(_x_lim is not None):
        x_min = _x_lim[0]
        x_max = _x_lim[1]
        if(len(_x_lim) > 2): nxInteg = int(round(_x_lim[2]))

    y_min = yStart; y_max = yEnd; nyInteg = 0
    if(_y_lim is not None):
        y_min = _y_lim[0]
        y_max = _y_lim[1]
        if(len(_y_lim) > 2): nyInteg = int(round(_y_lim[2]))

    if((nxInteg > 2) and (nyInteg > 2) and (_ar_align == 1)): #perform inregration using 2D interpolation
        
        arAuxIntegWave2Dx = array('d', [0]*nxInteg)
        if(x_min < xStart): x_min = xStart
        if(x_max > xEnd): x_max = xEnd
        xStepInteg = (x_max - x_min)/(nxInteg - 1)
        
        arAuxIntegWave2Dy = array('d', [0]*nyInteg)
        if(y_min < yStart): y_min = yStart
        if(y_max > yEnd): y_max = yEnd
        yStepInteg = (y_max - y_min)/(nyInteg - 1)

        yy = y_min
        for iy in range(nyInteg):
            xx = x_min
            for ix in range(nxInteg):
                arAuxIntegWave2Dx[ix] = interp_2d(xx, yy, xStart, xStep, nx, yStart, yStep, ny, _ar, _ord=2)
                xx += xStepInteg

            arAux = integ_array(arAuxIntegWave2Dx, xStepInteg)
            arAuxIntegWave2Dy[iy] = arAux[nxInteg - 1]
            yy += yStepInteg

        arAux = integ_array(arAuxIntegWave2Dy, yStepInteg)
        resInteg = arAux[nyInteg - 1]
        return resInteg

    ixStart = int((x_min - xStart)/xStep + 0.e-12)
    if(ixStart < 0): ixStart = 0
    else:
        if(ixStart >= nx): ixStart = nx - 1
    #print('ixStart=', ixStart)
    
    iyStart = int((y_min - yStart)/yStep + 0.e-12)
    if(iyStart < 0): iyStart = 0
    else:
        if(iyStart >= ny): iyStart = ny - 1
    #print('iyStart=', iyStart)

    ixEnd = int((x_max - xStart)/xStep - 1.e-06) + 1
    if(ixEnd < 0): ixEnd = 0
    else: 
        if(ixEnd >= nx): ixEnd = nx - 1
    #print('ixStart=', ixStart, 'ixEnd=', ixEnd)

    iyEnd = int((y_max - yStart)/yStep - 1.e-06) + 1
    if(iyEnd < 0): iyEnd = 0
    else:
        if(iyEnd >= ny): iyEnd = ny - 1
    #print('iyStart=', iyStart, 'iyEnd=', iyEnd)

    if((ixStart >= ixEnd) or (iyStart >= iyEnd)): return 0.

    nxi = ixEnd - ixStart + 1
    ##make/O/N=(nxi) wAuxIntegWave2Dx
    ##SetScale/P x, xStart + ixStart*xStep, xStep, "", wAuxIntegWave2Dx
    arAuxIntegWave2Dx = array('d', [0]*nxi)

    nyi = iyEnd - iyStart + 1
    ##make/O/N=(nyi) wAuxIntegWave2Dy
    ##SetScale/P x, yStart + iyStart*yStep, yStep, "", wAuxIntegWave2Dy
    arAuxIntegWave2Dy = array('d', [0]*nyi)

    ##for(iy = 0; iy < nyi; iy += 1)
    ##	wAuxIntegWave2Dx = w2d[ixStart + p][iyStart + iy]
    ##	integrate/T wAuxIntegWave2Dx
    ##	wAuxIntegWave2Dy[iy] = wAuxIntegWave2Dx(x_max) - wAuxIntegWave2Dx(x_min)
    ##endfor

    iyAbs_nx = 0
    ar_iyAbs = None
    for iy in range(nyi):
        iyAbs = iy + iyStart
        if(_ar_align == 1): iyAbs_nx = iyAbs*nx
        else: ar_iyAbs = _ar[iyAbs]
        
        for ix in range(nxi):
            ixAbs = ix + ixStart
            if(_ar_align == 1): arAuxIntegWave2Dx[ix] = _ar[iyAbs_nx + ixAbs]
            else: arAuxIntegWave2Dx[ix] = ar_iyAbs[ixAbs]

        #print(xStep, arAuxIntegWave2Dx)
        arAux = integ_array(arAuxIntegWave2Dx, xStep)
        arAuxIntegWave2Dy[iy] = arAux[nxi - 1]

    ##integrate/T wAuxIntegWave2Dy
    ##variable res = wAuxIntegWave2Dy(y_max) - wAuxIntegWave2Dy(y_min)

    arAux = integ_array(arAuxIntegWave2Dy, yStep)
    resInteg = arAux[nyi - 1]
    return resInteg

#****************************************************************************
def matr_prod(_A, _B):
    """
    Multiplies matrix _A by matrix or by vector _B 
    """
    # Matrix multiplication
    B0 = _B[0]
    lenB = len(_B)
    lenA = len(_A)
    if(len(_A[0]) != lenB): # Check matrix dimensions        
        Exception('Matrices have wrong dimensions')
    if(isinstance(B0, list) or isinstance(B0, array) or isinstance(B0, tuple)): #_B is matrix
        lenB0 = len(B0)
        C = [[0 for row in range(lenB0)] for col in range(lenA)]
        for i in range(lenA):
            for j in range(lenB0):
                for k in range(lenB):
                    C[i][j] += _A[i][k]*_B[k][j]
    else: #_B is vector
        C = [0 for row in range(lenB)]
        for i in range(lenA):
            for k in range(lenB):
                C[i] += _A[i][k]*_B[k]
    return C

#****************************************************************************
def matr_transp(_A):
    """
    Returns transposed matrix
    """
    lenA = len(_A)
    lenA0 = len(_A[0])
    return [[_A[i][j] for i in range(lenA)] for j in range(lenA0)]

#****************************************************************************
def matr_print(_A):
    """
    Prints matrix _A
    """
    for i in range(len(_A)):
        print(_A[i])

#****************************************************************************
def matr3x3_det(_M):
    S0 = _M[0]; S1 = _M[1]; S2 = _M[2]
    return S0[0]*S1[1]*S2[2] + S0[1]*S1[2]*S2[0] + S0[2]*S1[0]*S2[1] - S0[2]*S1[1]*S2[0] - S0[0]*S1[2]*S2[1] - S0[1]*S1[0]*S2[2]

def matr_3x3_det(_M):
    return matr3x3_det(_M)

#****************************************************************************
def matr3x3_inv(_M):
    S0 = _M[0]; S1 = _M[1]; S2 = _M[2]
    invDet = 1./(S0[0]*S1[1]*S2[2] + S0[1]*S1[2]*S2[0] + S0[2]*S1[0]*S2[1] - S0[2]*S1[1]*S2[0] - S0[0]*S1[2]*S2[1] - S0[1]*S1[0]*S2[2])
    S0i = [invDet*(S1[1]*S2[2] - S1[2]*S2[1]), invDet*(-S0[1]*S2[2] + S0[2]*S2[1]), invDet*(S0[1]*S1[2] - S0[2]*S1[1])]
    S1i = [invDet*(-S1[0]*S2[2] + S1[2]*S2[0]), invDet*(S0[0]*S2[2] - S0[2]*S2[0]), invDet*(-S0[0]*S1[2] + S0[2]*S1[0])]
    S2i = [invDet*(S1[0]*S2[1] - S1[1]*S2[0]), invDet*(-S0[0]*S2[1] + S0[1]*S2[0]), invDet*(S0[0]*S1[1] - S0[1]*S1[0])]
    return [S0i, S1i, S2i]

def matr_3x3_inv(_M):
    return matr3x3_inv(_M)

#****************************************************************************
def vect_prod_s(_V1, _V2):
    """
    Returns scalar product of vectors V1 and V2
    """
    sizeV1 = len(_V1)
    sizeV2 = len(_V2)
    sizeV = sizeV1 if(sizeV1 < sizeV2) else sizeV2
    res = 0
    for i in range(sizeV): res += _V1[i]*_V2[i]
    return res

#****************************************************************************
def vect3_prod_v(_V1, _V2):
    """
    Returns vector product of 3d vectors V1 and V2
    """
    return [_V1[1]*_V2[2] - _V1[2]*_V2[1], _V1[2]*_V2[0] - _V1[0]*_V2[2], _V1[0]*_V2[1] - _V1[1]*_V2[0]]

#****************************************************************************
def vect_norm(_V):
    """
    Returns vector norm (/length)
    """
    return sqrt(sum(n**2 for n in _V))

#****************************************************************************
def vect_normalize(_V):
    """
    Normalizes vector (in place, i.e. without creating new vector)
    """
    invNorm = 1./vect_norm(_V)
    for i in range(len(_V)): _V[i] *= invNorm
    return _V

#****************************************************************************
def vect_mult(_V, _a):
    """
    Multiplies vector _V (in place) by number _a
    """
    for i in range(len(_V)): _V[i] *= _a
    return _V

#****************************************************************************
def trf_rotation(_V, _ang, _P):
    """
    Sets up matrix and vector describing rotation about axis _V passing through a point _P about an angle _ang
    :param _V: vector (array of 3 Cartesian coordinates) defining rotation axis
    :param _ang: rotation angle [rad]
    :param _P: point (array of 3 Cartesian coordinates) rotation axis passes through
    :returns list containing the 3x3 matrix and 3-element vector
    """
    normFact = 1./sqrt(_V[0]*_V[0] + _V[1]*_V[1] + _V[2]*_V[2])
    axVect = [normFact*_V[0], normFact*_V[1], normFact*_V[2]]
    VxVx = axVect[0]*axVect[0]
    VyVy = axVect[1]*axVect[1]
    VzVz = axVect[2]*axVect[2]
    cosAng = cos(_ang)
    sinAng = sin(_ang)
    one_m_cos = 1. - cosAng
    one_m_cosVxVy = one_m_cos*axVect[0]*axVect[1]
    one_m_cosVxVz = one_m_cos*axVect[0]*axVect[2]
    one_m_cosVyVz = one_m_cos*axVect[1]*axVect[2]
    sinVx = sinAng*axVect[0]
    sinVy = sinAng*axVect[1]
    sinVz = sinAng*axVect[2]
    st0 = [VxVx + cosAng*(VyVy + VzVz), one_m_cosVxVy - sinVz, one_m_cosVxVz + sinVy]
    st1 = [one_m_cosVxVy + sinVz, VyVy + cosAng*(VxVx + VzVz), one_m_cosVyVz - sinVx]
    st2 = [one_m_cosVxVz - sinVy, one_m_cosVyVz + sinVx, VzVz + cosAng*(VxVx + VyVy)]
    M = [st0, st1, st2]
    st00 = [1. - st0[0], -st0[1], -st0[2]]
    st01 = [-st1[0], 1. - st1[1], -st1[2]]
    st02 = [-st2[0], -st2[0], 1. - st2[2]]
    M0 = [st00, st01, st02]
    V = matr_prod(M0, _P)
    return [M, V]

#****************************************************************************
def fwhm(x, y, shift=0.5, return_as_dict=False):  # MR21032017
    """The function searches x-values (roots) where y=0 (after normalization to values between 0 and 1 and shifting the
    values down by 0.5 (default value)) based on linear interpolation, and calculates full width at half maximum (FWHM).

    :param x: an array of x values.
    :param y: an array of y values.
    :param shift: an optional shift to be used in the process of normalization (between 0 and 1).
    :param return_as_dict: if to return a dict with 'fwhm' and 'x_range'
    :return: a value of the FWHM or dictionary consisting of 'fwhm' and 'x_range'
    """

    def is_positive(num):
        return True if num > 0 else False

    # Normalize values first:
    #y = (y - min(y)) / (max(y) - min(y)) - shift  # roots are at Y=0

    #OC18112017 (making it work with standard Python lists / arrays)
    minY = min(y)
    maxY = max(y)
    if(maxY == minY): raise Exception('FWHM can not be calculated')
    mult = 1./(maxY - minY)
    lenY = len(y)
    for i in range(lenY): y[i] = (y[i] - minY)*mult - shift

    positive = is_positive(y[0])
    list_of_roots = []
    #for i in range(len(y)):
    for i in range(lenY):
        current_positive = is_positive(y[i])
        if current_positive != positive:
            list_of_roots.append(x[i - 1] + (x[i] - x[i - 1]) / (abs(y[i]) + abs(y[i - 1])) * abs(y[i - 1]))
            positive = not positive
    if len(list_of_roots) >= 2:
        if not return_as_dict:
            return abs(list_of_roots[-1] - list_of_roots[0])
        else:
            return {
                'fwhm': abs(list_of_roots[-1] - list_of_roots[0]),
                'x_range': list_of_roots,
            }
    else:
        raise Exception('Number of roots is less than 2!')

#****************************************************************************
#def fwhm_scipy(x, y): #MR27092016
#    """Computing FWHM (Full width at half maximum)"""
#    try:
#        from scipy.interpolate import UnivariateSpline
#        spline = UnivariateSpline(x, y, s=0)
#        r1, r2 = spline.roots()  # find the roots
#        return r2 - r1  # return the difference (full width)
#    except ImportError:
#        return fwhm(x, y)

#****************************************************************************
def get_dist_uni(_min, _max): #RAC30032020
    '''Select point using a uniform distribution

    :param _min: minimum possible value.
    :param _max: maximum possible value.
    '''
    ### Load package Numpy
    #try:
    #    import numpy as np
    #except:
    #    print('NumPy can not be loaded. You may need to install numpy. If you are using pip, you can use the following ' + 
    #          "command to install it: \npip install numpy")
        
    #if int(_min) == int(_max):
    #    return _max
    #else:
    #    return np.random.randint(_min, _max)
    
    return random.uniform(_min, _max) #OC24052020

#****************************************************************************
#def get_dist_norm(_min, _max, _scale=1.0): #OC24052020
def get_dist_norm(_min, _max, _scale=1.0, _size=None): #RAC30032020
    '''Select point using a normal (Gaussian) distribution

    :param _min: minimum possible value.
    :param _max: maximum possible value.
    :param _scale = 1.0: Standard deviation (spread or “width”) of the distribution.
    :param _size = None: (int or tuple of ints) 
        Output shape. If the given shape is, e.g., (m, n, k), then m * n * k 
        samples are drawn. If size is None (default), a single value is  returned 
        if loc and scale are both scalars. Otherwise, np.broadcast(loc, scale).size 
        samples are drawn.

    The probability density for the Gaussian distribution is:

    p(x) = \frac{1}{\sqrt{ 2 \pi \sigma^2 }} e^{ - \frac{ (x - \mu)^2 } {2 \sigma^2} }

    Where \mu is the mean and \sigma the standard deviation. The square of the 
    standard deviation, \sigma^2, is called the variance. The function has its peak 
    at the mean, and its “spread” increases with the standard deviation 
    (the function reaches 0.607 times its maximum at x + \sigma and x - \sigma [2]). 
    This implies that numpy.random.normal is more likely to return samples lying 
    close to the mean, rather than those far away.
    '''
    ### Load package Numpy
    #try:
    #    import numpy as np
    #except:
    #    print('NumPy can not be loaded. You may need to install numpy. If you are using pip, you can use the following ' + 
    #          "command to install it: \npip install numpy")
    #try:
    #    _max >= _min
    #except:
    #    print("Maximum distribution value is less than minimum distribution value")

    #Get mean (“centre location”) of the distribution.
    _loc = (_max+_min)/2 #OC24052020
    #_loc = (_max-_min)/2
    #print('_min=', _min, '_max=', _max, '_scale=', _scale) #DEBUG

    #Get intiger from normal (Gaussian) distribution centered at _loc
    _norm_var = 0
    while _norm_var <= _min or _norm_var >= _max:
        #OC24052020
        _norm_var = random.gauss(_loc, _scale) #scale=_scale, size=_size) 
        #_norm_var = int(np.random.normal(loc=_loc, scale=_scale, size=_size))
        #print('   _norm_var=', _norm_var) #DEBUG

    return _norm_var

#****************************************************************************
def get_dist_schultz(_min, _max, poly_index=0.3): #RAC30032020
    '''elect point using a Flory–Schulz distribution

    :param _min: minimum possible value.
    :param _max: maximum possible value.
    :param poly_index = 0.3: particles of varied sizes in the dispersed 
        phase of a disperse system.

    The Flory–Schulz distribution is a discrete probability distribution named 
    after Paul Flory and Günter Victor Schulz that describes the relative ratios 
    of polymers of different length that occur in an ideal step-growth 
    polymerization process. 

    Calculate the Schulz distribution for polydisperse systems. When the
    polydispersity is small, the Schulz distribution tends to a Gaussian. 

    Flory–Schulz Distribution:
        y = y(0) + Ae^((x(c)-x)/w) * (x/x(c))^(x(c)/w)

    Where:
        y(0) = minimum (_min).
        A = amplitude (_max-_min).
        x(c) = center of distribution (mean of x_range).
        w = width of distribution ((1 - poly_index^2) / (poly_index^2)).
    '''
    ### Load package Numpy
    try:
        import numpy as np
    except:
        print('NumPy can not be loaded. You may need to install numpy. If you are using pip, you can use the following ' + 
              "command to install it: \npip install numpy")

    #'Width' of the distribution
    width = int((1 - np.power(poly_index, 2)) / np.power(poly_index, 2))

    #Get a range of values
    x_range = np.linspace(1,500)

    #Get Mean of Distribution
    mean = np.mean(x_range)

    #Get Schulz Distribution
    schulz = (_min + (_max-_min) * (np.exp((mean-x_range)/width)) * np.power((x_range/mean),(mean/width)))

    return int(np.random.choice(schulz))
